import unittest
import numpy as np
from hw2d.initializations.fourier_noise import get_fft_noise


class TestFFTNoise(unittest.TestCase):
    def test_fft_noise_output_shape(self):
        resolution = (256, 256)
        size = 1.0
        scale = 2.0
        noise = get_fft_noise(resolution, size, scale)

        # Test that output is numpy array
        self.assertIsInstance(noise, np.ndarray)

        # Test the output shape matches the given resolution
        self.assertEqual(noise.shape, resolution)

    def test_fft_noise_mean_std(self):
        resolution = (256, 256)
        size = 1.0
        scale = 2.0
        noise = get_fft_noise(resolution, size, scale)

        # Check if the noise mean is close to zero and standard deviation is close to 1
        self.assertTrue(np.isclose(np.mean(noise), 0.0, atol=1e-5))
        self.assertTrue(np.isclose(np.std(noise), 1.0, atol=1e-5))


if __name__ == "__main__":
    unittest.main()



from typing import Tuple
import numpy as np


def get_2d_sine(grid_size: Tuple[int, int], L: float) -> np.ndarray:
    """
    Generate a 2D sine pattern over a specified grid.

    This function creates a 2D array where the values are determined by the sine
    function applied to both the x and y coordinates. The resulting pattern can be
    visualized as a 2D sine wave in both the x and y directions.

    Args:
        grid_size (Tuple[int, int]): Size of the grid (height, width) for the 2D sine pattern.
        L (float): Physical extent of the domain. The resulting values will be in the range [0, L).

    Returns:
        np.ndarray: A 2D numpy array of shape `grid_size` containing the sine pattern.

    Example:
        >>> get_2d_sine((5,5), 1.0)
        array([[ 0.        ,  0.58778525,  0.95105652,  0.95105652,  0.58778525],
               [ 0.        ,  0.48163507,  0.78183148,  0.78183148,  0.48163507],
               [-0.        , -0.48163507, -0.78183148, -0.78183148, -0.48163507],
               [-0.        , -0.58778525, -0.95105652, -0.95105652, -0.58778525],
               [-0.        , -0.48163507, -0.78183148, -0.78183148, -0.48163507]])
    """
    indices = np.array(np.meshgrid(*list(map(range, grid_size))))
    phys_coord = indices.T * L / (grid_size[0])  # between [0, L)
    x, y = phys_coord.T
    d = np.sin(2 * np.pi * x + 1) * np.sin(2 * np.pi * y + 1)
    return d



class Namespace(dict):
    def __mul__(self, other):
        if isinstance(other, Namespace):
            return Namespace({key: other[key] * val for key, val in self.items()})
        else:
            return Namespace({key: other * val for key, val in self.items()})

    __rmul__ = __mul__

    def __div__(self, other):
        if isinstance(other, Namespace):
            return Namespace({key: val / other[key] for key, val in self.items()})
        else:
            return Namespace({key: val / other for key, val in self.items()})

    def __truediv__(self, other):
        if isinstance(other, Namespace):
            return Namespace({key: val / other[key] for key, val in self.items()})
        else:
            return Namespace({key: val / other for key, val in self.items()})

    def __rdiv__(self, other):
        if isinstance(other, Namespace):
            return Namespace({key: other[key] / val for key, val in self.items()})
        else:
            return Namespace({key: other / val for key, val in self.items()})

    def __add__(self, other):
        if isinstance(other, Namespace):
            return Namespace({key: other[key] + val for key, val in self.items()})
        else:
            return Namespace({key: other + val for key, val in self.items()})

    __radd__ = __add__

    def __sub__(self, other):
        if isinstance(other, Namespace):
            return Namespace({key: other[key] - val for key, val in self.items()})
        else:
            return Namespace({key: other - val for key, val in self.items()})

    @property
    def omega(self):
        return self["omega"]

    @property
    def phi(self):
        return self["phi"]

    @property
    def density(self):
        return self["density"]

    @property
    def dE(self):
        return self["dE"]

    @property
    def dU(self):
        return self["dU"]

    @property
    def age(self):
        return self["age"]

    @property
    def dtype(self):
        return self["density"].dtype

    def copy(self):
        return Namespace({key: val for key, val in self.items()})



import fire
import numpy as np
import matplotlib.pyplot as plt
from tqdm import tqdm
import os
from typing import Iterable

# Local imports
from hw2d.model import HW
from hw2d.initializations.fourier_noise import get_fft_noise
from hw2d.initializations.sine import get_2d_sine
from hw2d.utils.io import (
    get_save_params,
    create_appendable_h5,
    save_to_buffered_h5,
    append_h5,
    continue_h5_file,
)
from hw2d.utils.namespaces import Namespace
from hw2d.utils.plot.movie import create_movie
from hw2d.utils.run_properties import calculate_properties
from hw2d.utils.plot.timetrace import plot_timetraces
from hw2d.utils.downsample import fourier_downsample


def run(
    # Physics & Numerics
    step_size: float = 0.025,
    end_time: float = 300,
    grid_pts: int = 64,
    k0: float = 0.15,
    N: int = 3,
    nu: float = 5.0e-08,
    c1: float = 1.0,
    kappa_coeff: float = 1.0,
    poisson_bracket_coeff: float = 1.0,
    # Initialization
    seed: int or None = None,
    init_type: str = "normal",
    init_scale: float = 1 / 100,
    # Saving
    output_path: str = "_test.h5",
    continue_file: bool = False,
    buffer_length: int = 100,
    snaps: int = 1,
    downsample_factor: float = 2,
    # Movie
    movie: bool = True,
    min_fps: int = 10,
    dpi: int = 75,
    speed: int = 5,
    # Properties
    properties: Iterable[str] = [
        "gamma_n",
        "gamma_n_spectral",
        "gamma_c",
        "energy",
        "thermal_energy",
        "kinetic_energy",
        "enstrophy",
        "enstrophy_phi",
    ],
    # Plotting
    plot_properties: Iterable[str] = (
        "enstrophy",
        "energy",
        "kinetic_energy",
        "thermal_energy",
    ),
    # Other
    debug: bool = False,
):
    """
    Run the simulation with the given parameters.

    Args:
        step_size (float, optional): Incremental step for simulation progression. Defaults to 0.025.
        end_time (float, optional): Duration till the simulation should run. Defaults to 1_000.
        grid_pts (int, optional): Grid points. Suggested: 128 for coarse, 1024 for fine. Defaults to 512.
        k0 (float, optional): Determines k-focus. Suggested: 0.15 for high-k, 0.0375 for low-k. Defaults to 0.15.
        N (int, optional): Dissipation exponent's half value. Defaults to 3.
        nu (float, optional): Viscosity. Suggested: 5e-10 for coarse-large, 1e-4 for fine-small. Defaults to 5.0e-08.
        c1 (float, optional): Transition scale between hydrodynamic and adiabatic. Suggested values: 0.1, 1, 5. Defaults to 1.0.
        kappa_coeff (float, optional): Coefficient of d/dy phi. Defaults to 1.0.
        poisson_bracket_coeff (float, optional): Coefficient of Poisson bracket [A,B] implemented with Arakawa Scheme. Defaults to 1.0.
        seed (int or None, optional): Seed for random number generation. Defaults to None.
        init_type (str, optional): Initialization method. Choices: 'fourier', 'sine', 'random', 'normal'. Defaults to 'normal'.
        init_scale (float, optional): Scaling factor for initialization. Defaults to 0.01.
        output_path (str, optional): Where to save the simulation data. Defaults to ''.
        continue_file (bool, optional): If True, continue with existing file. Defaults to False.
        buffer_length (int, optional): Size of buffer for storage. Defaults to 100.
        snaps (int, optional): Snapshot intervals for saving. Defaults to 1.
        movie (bool, optional): If True, generate a movie out of simulation. Defaults to True.
        min_fps (int, optional): Parameter for movie generation. Defaults to 10.
        dpi (int, optional): Parameter for movie generation. Defaults to 75.
        speed (int, optional): Parameter for movie generation. Defaults to 5.
        properties (Iterable[str], optional): List of properties to calculate for the saved file.
        plot_properties (Iterable[str], optional): List of properties to plot a timetrace for.
        debug (bool, optional): Enable or disable debug mode. Defaults to False.

    Returns:
        None: The function saves simulation data or generates a movie as specified.
    """

    # Unpacking
    y = grid_pts
    x = grid_pts
    L = 2 * np.pi / k0  # Physical box size
    dx = L / x  # Grid resolution
    steps = int(end_time / step_size)  # Number of Steps until end_time
    snap_count = steps // snaps + 1  # number of snapshots
    field_list = ("density", "omega", "phi")
    current_time = 0
    iteration_count = 0
    np.random.seed(seed)

    # Define Initializations
    noise = {
        "fourier": lambda y, x: get_fft_noise(
            resolution=[y, x],
            size=L,
            scale=1,
            min_wavelength=dx * 12,
            max_wavelength=dx * grid_pts * 100,
            factor=2,
        ),
        "sine": lambda y, x: get_2d_sine((y, x), L),
        "random": lambda y, x: np.random.rand(y, x).astype(np.float64),
        "normal": lambda y, x: np.random.normal(size=(y, x)).astype(np.float64),
    }
    downsample_fnc = fourier_downsample

    # Physics
    physics_params = dict(
        dx=dx,
        N=N,
        c1=c1,
        nu=nu,
        k0=k0,
        poisson_bracket_coeff=poisson_bracket_coeff,
        kappa_coeff=kappa_coeff,
    )
    # Initialize Plasma
    plasma = Namespace(
        density=noise[init_type](y, x) * init_scale,
        omega=noise[init_type](y, x) * init_scale,
        phi=noise[init_type](y, x) * init_scale,
        age=0,
    )

    # File Handling
    if continue_file:
        # Continue from previous run
        plasma, physics_params = continue_h5_file(continue_file, field_list)
        current_time = plasma.age
        # Check if broken
        for field in plasma.keys():
            if np.isnan(np.sum(plasma[field])):
                print(f"FAILED @ {iteration_count:,} steps ({plasma.age:,})")
                raise BaseException(f"Input File is broken: FAILED @ {iteration_count:,} steps ({plasma.age:,})")
        print(
            f"Successfully loaded: {output_path} (age={plasma.age})\n{physics_params}"
        )
    if output_path:
        y_save = y
        x_save = x
        if downsample_factor != 1:
            y_save = int(round(y / downsample_factor))
            x_save = int(round(x / downsample_factor))
            print(f"Downsampling by a factor {downsample_factor} for saving to: {y_save}x{x_save}")
        # Output data from this run
        buffer = {
            field: np.zeros((buffer_length, y_save, x_save), dtype=np.float32)
            for field in field_list
        }
        output_params = {
            "buffer": buffer,
            "buffer_index": 0,
            "output_path": output_path,
        }
        # Load Data
        if os.path.isfile(output_path):
            if not continue_file:
                print(f"File already exists.")
                return
        # Create
        else:
            save_params = get_save_params(physics_params, step_size, snaps, x, y, x_save=x_save, y_save=y_save)
            create_appendable_h5(
                output_path, save_params, dtype=np.float32, chunk_size=100
            )
            new_val = plasma
            if downsample_factor != 1:
                new_val = Namespace(**{k: downsample_fnc(v, downsample_factor) for k,v in plasma.items() if k in ("phi", "omega", "density")})
            output_params["buffer_index"] = save_to_buffered_h5(
                new_val=new_val, buffer_length=buffer_length, **output_params
            )

    # Setup Simulation
    hw = HW(**physics_params, debug=debug)
    plasma["phi"] = hw.get_phi(plasma.omega, physics_params["dx"])

    # Run Simulation
    print("Running simulation...")
    for i in tqdm(range(1, steps + 1)):
        # Progress one step, alternatively: hw.euler_step()
        plasma = hw.rk4_step(plasma, dt=step_size, dx=dx)

        # Save to records
        if output_path and i % snaps == 0:
            new_val = plasma
            if downsample_factor != 1:
                new_val = Namespace(**{k: downsample_fnc(v, downsample_factor) for k,v in plasma.items() if k in ("phi", "omega", "density")})
            output_params["buffer_index"] = save_to_buffered_h5(
                new_val=new_val, buffer_length=buffer_length, **output_params
            )

        # Check for breaking
        if np.isnan(np.sum(plasma.density)):
            print(f"FAILED @ {i:,} steps ({plasma.age:,})")
            break

    # If output_path is defined, flush any remaining data in the buffer
    if output_path and output_params["buffer_index"] > 0:
        append_h5(**output_params)

    # Get Performance stats
    hw.print_log()

    # Generate Movie from saved file
    if movie and output_path:
        print(f"Generating movie...")
        create_movie(
            input_filename=output_path,
            output_filename=output_path.replace(".h5", ""),
            t0=0,
            t1=None,
            plot_order=field_list,
            min_fps=min_fps,
            dpi=dpi,
            speed=speed,
        )

    if properties and output_path:
        print(f"Calculating properties...")
        calculate_properties(
            file_path=output_path,
            batch_size=buffer_length,
            property_list=properties,
            force_recompute=True,
            is_debug=False,
        )

    if plot_properties and output_path:
        print(f"Plotting properties...")
        plot_timetraces(
            file_path=output_path,
            out_path=None,
            properties=plot_properties,
            t0=0,
            t0_std=300,
        )


if __name__ == "__main__":
    fire.Fire(run)



import numpy as np
import perfplot as pp
import fire
from typing import List


# Function definitions
def setup_random_array(n, scale=1000):
    return np.random.rand(n, n) / scale

def setup_random_arrays_pair(n, scale=1000):
    return [setup_random_array(n, scale), setup_random_array(n, scale)]

def get_function_defs(function_name):
    if function_name == "periodic_gradient":
        from hw2d.gradients.numpy_gradients import periodic_gradient as np_periodic_gradient
        from hw2d.gradients.numba_gradients import periodic_gradient as nb_periodic_gradient
        np_instrinsic_gradient = lambda x, dx: np.gradient(np.pad(x, 1, mode="wrap"), axis=-2)[1:-1, 1:-1]
        return [
            ("np_periodic_gradient", np_periodic_gradient),
            ("np_instrinsic_gradient", np_instrinsic_gradient),
            ("nb_periodic_gradient", nb_periodic_gradient),
        ]
    elif function_name == "periodic_laplace":
        from hw2d.gradients.numpy_gradients import periodic_laplace as np_periodic_laplace_N
        from hw2d.gradients.numba_gradients import periodic_laplace as nb_periodic_laplace_N
        from hw2d.gradients.numpy_gradients import fourier_laplace as np_fourier_laplace
        return [
            ("np_periodic_laplace", np_periodic_laplace_N),
            #("np_fourier_laplace", np_fourier_laplace),
            ("nb_periodic_laplace", nb_periodic_laplace_N),
        ]
    elif function_name == "poisson_solvers":
        from hw2d.poisson_solvers.numpy_fourier_poisson import fourier_poisson_double as np_fourier_poisson_double
        from hw2d.poisson_solvers.numba_fourier_poisson import fourier_poisson_double as nb_fourier_poisson_double
        return [
            ("np_fourier_poisson_double", np_fourier_poisson_double),
            ("nb_fourier_poisson_double", nb_fourier_poisson_double),
        ]
    elif function_name == "poisson_bracket":
        from hw2d.poisson_bracket.numpy_arakawa import periodic_arakawa as np_arakawa
        from hw2d.poisson_bracket.numpy_arakawa import periodic_arakawa_vec as np_arakawa_vec
        from hw2d.poisson_bracket.numba_arakawa import periodic_arakawa as nb_arakawa
        from hw2d.poisson_bracket.numba_arakawa import periodic_arakawa_stencil as nb_arakawa_stencil
        from hw2d.poisson_bracket.numba_arakawa import periodic_arakawa_vec as nb_arakawa_vec
        return [
            ("np_arakawa", np_arakawa),
            ("np_arakawa_vec", np_arakawa_vec),
            ("nb_arakawa", nb_arakawa),
            ("nb_arakawa_stencil", nb_arakawa_stencil),
            ("nb_arakawa_vec", nb_arakawa_vec)
        ]

# Main function
def main(
    max_time: int = 30,  # seconds
    scale: int = 1000,
    dx: float = 1,
    function_names: List[str] = ["periodic_gradient", "periodic_laplace", "poisson_bracket", "poisson_solvers"],  # Add or remove function names as needed
):
    sizes = 2 ** np.array(range(4, 13))
    for function_name in function_names:
        # Get function definitions
        function_defs = get_function_defs(function_name)
        # Repack to labels and kernels
        labels = [t[0] for t in function_defs]
        kernels = [t[1] for t in function_defs]
        # Setup functions for benchmark
        if function_name != "poisson_bracket":
            kernels = [lambda a, f=f: f(a, dx=dx) for f in kernels]
            setup_fnc = setup_random_array
        else:
            kernels = [lambda ab, f=f: f(*ab, dx=dx) for f in kernels]
            setup_fnc = setup_random_arrays_pair
        # Bechmark
        benchmark = pp.bench(
            setup=setup_fnc,
            kernels=kernels,
            labels=labels,
            n_range=sizes,
            xlabel="len(a)",
            max_time=max_time,
        )
        # Plot
        #benchmark.show()
        print(benchmark)

if __name__ == "__main__":
    fire.Fire(main)



import numpy as np
from numba import jit


@jit(nopython=True, nogil=True, parallel=True)
def laplace(padded: np.ndarray, dx: float) -> np.ndarray:
    """
    Compute the gradient of a 2D array using finite differences with Numba optimization.

    Args:
        padded (np.ndarray): 2D array with padding of size 1.
        dx (float): The spacing between grid points.

    Returns:
        np.ndarray: The gradient of the input array.
    """
    return (
        padded[0:-2, 1:-1]  # above
        + padded[1:-1, 0:-2]  # left
        - 4 * padded[1:-1, 1:-1]  # center
        + padded[1:-1, 2:]  # right
        + padded[2:, 1:-1]  # below
    ) / dx**2


def periodic_laplace(arr: np.ndarray, dx: float) -> np.ndarray:
    """
    Compute the gradient of a 2D array using finite differences with periodic boundary conditions and Numba optimization.

    Args:
        a (np.ndarray): Input 2D array.
        dx (float): The spacing between grid points.

    Returns:
        np.ndarray: The gradient of the input array with periodic boundary conditions.
    """
    return laplace(np.pad(arr, 1, "wrap"), dx)


@jit(nopython=True)
def laplace_N(arr: np.ndarray, dx: float, N: int) -> np.ndarray:
    """
    Compute the gradient of a 2D array N times successively.

    Args:
        arr (np.ndarray): Input 2D array.
        dx (float): The spacing between grid points.
        N (int): Number of iterations.

    Returns:
        np.ndarray: The gradient of the input array after N iterations.
    """
    for _ in range(N):
        arr = laplace(arr, dx)
    return arr


def periodic_laplace_N(arr: np.ndarray, dx: float, N: int) -> np.ndarray:
    """
    Compute the gradient of a 2D array N times successively with periodic boundary conditions.

    Args:
        arr (np.ndarray): Input 2D array.
        dx (float): The spacing between grid points.
        N (int): Number of iterations.

    Returns:
        np.ndarray: The gradient of the input array after N iterations with periodic boundary conditions.
    """
    return laplace_N(np.pad(arr, N, "wrap"), dx, N)


@jit(nopython=True)
def gradient(padded: np.ndarray, dx: float, axis: int = 0) -> np.ndarray:
    """
    Compute the gradient of a 2D array using finite differences with Numba optimization.

    Args:
        padded (np.ndarray): 2D array with padding of size 1.
        dx (float): The spacing between grid points.
        axis (int): Axis along which the gradient is tkaen

    Returns:
        np.ndarray: Gradient in axis-direction.
    """
    if axis == 0:
        return (padded[2:, 1:-1] - padded[0:-2, 1:-1]) / (2 * dx)
    elif axis == 1:
        return (padded[1:-1, 2:] - padded[1:-1, 0:-2]) / (2 * dx)


def periodic_gradient(input_field: np.ndarray, dx: float, axis: int = 0) -> np.ndarray:
    """
    Compute the gradient of a 2D array using finite differences with periodic boundary condition with Numba optimization

    Args:
        padded (np.ndarray): 2D array with padding of size 1.
        dx (float): The spacing between grid points.
        axis (int): Axis along which the gradient is tkaen

    Returns:
        np.ndarray: Gradient in axis-direction with periodic boundary conditions.
    """
    padded = np.pad(input_field, 1, mode="wrap")
    return gradient(padded, dx, axis=axis)



import fire
import numpy as np
import h5py
from tqdm import tqdm
from typing import Dict, Union, List, Tuple
from matplotlib.colors import Colormap
import matplotlib.pyplot as plt

from hw2d.utils.plot.movie import get_extended_viridis
from hw2d.utils.latex_format import latex_format


def plot_dict(
    plot_dic: Dict,
    cmap: Union[str, Colormap, List[str]],
    couple_cbars: bool = False,
    figsize: Tuple or None = None,
    sharex: bool = False,
    sharey: bool = False,
    vertical: bool = False,
    equal: bool = True,
    cbar_label_spacing: float or None = None,
):
    n = len(plot_dic.values())
    labels = list(plot_dic.keys())
    if figsize is None:
        figsize = (n * 4, 4)
    if vertical:
        fig, axarr = plt.subplots(n, 1, figsize=figsize, sharex=sharex)
    else:
        fig, axarr = plt.subplots(
            1, n, figsize=figsize, sharey=sharey, constrained_layout=True
        )
    vmin = 0
    vmax = 0
    imgs = []
    cbars = []
    for i in tqdm(range(n)):
        # Local Names
        label = labels[i]
        data = plot_dic[label]
        ax = axarr[i]
        if len(data.shape) == 1:
            ax.axhline(0, color="k", alpha=0.7, linewidth=1)
            ax.plot(data)
        elif len(data.shape) == 2:
            imgs.append(ax.imshow(data, cmap=cmap))
            cbars.append(plt.colorbar(imgs[-1], pad=0, fraction=0.05))
            cax = cbars[-1].ax
            if cbar_label_spacing is not None:
                cax.yaxis.set_tick_params(pad=cbar_label_spacing)
            cax.xaxis.set_tick_params(pad=0)
        else:
            raise BaseException(
                f"Shape not recognized.  {label} has shape {data.shape}"
            )
        # Properties
        ax.set_title(label, pad=4)
        vmin = min(np.min(data), vmin)
        vmax = max(np.max(data), vmax)
        if equal:
            ax.set_aspect("equal")
    # fix colorbars together
    if couple_cbars:
        for img in imgs:
            img.set_clim(vmin, vmax)
    return fig


def main(
    file_path: str = "_512.h5",
    t: float = 200,
    save_type: str = "jpg",
    zero_omega: bool = True,
    # save_path: str = "imgs/raw/gridsize=512x512_k0=0.15_c1=1.0_dt=0.025_arak=1_kappa=1_N=3_nu=5.0e-06_scale=0.01_age=1000.000_version=6.pdf",
    ticks: bool = False,
):
    plot_order = ["density", "omega", "phi"]
    save_file_path = file_path.replace(".h5", ".jpg") + f"_t={t}"

    # Setting up paths and stuff
    if isinstance(t, float) and t.is_integer():
        t = int(t)

    file = h5py.File(file_path, "r")
    parameters = dict(file.attrs)

    t_idx = int(t // parameters["frame_dt"])
    print(t_idx)

    # Structure data
    plot_dic = {field_name: file[field_name][t_idx] for field_name in plot_order}
    x, y = plot_dic[plot_order[0]].shape

    # Center Omega?
    if zero_omega:
        plot_dic["omega"] = plot_dic["omega"] - np.mean(plot_dic["omega"])

    # Name Properly
    plot_dic = {latex_format(name): value for name, value in plot_dic.items()}

    ve = get_extended_viridis(vals=600)

    # Plot
    fig = plot_dict(
        plot_dic,
        cmap=ve,
        couple_cbars=False,
        figsize=None,
        sharex=True,
        sharey=True,
        vertical=False,
        cbar_label_spacing=1.7,
    )

    for i, ax in enumerate(fig.get_axes()):
        if i >= len(plot_order):
            break
        if ticks:
            ax.set_xticks([0, x // 2, x])
            if i == 0:
                ax.set_yticks([0, y // 2, y])
        else:
            ax.set_xticks([])
            ax.set_yticks([])

    # Adjust engine again
    fig.get_layout_engine().set(w_pad=0.013, h_pad=0.01, hspace=0, wspace=0)

    if save_file_path:
        fig.savefig(save_file_path[:-4] + ".jpg")
        fig.savefig(save_file_path[:-4] + ".pdf")
    print(save_file_path)


if __name__ == "__main__":
    main()



import numpy as np
import scipy.signal
import pytest
from unittest import TestCase
from hw2d.initializations.sine import get_2d_sine
from hw2d.gradients.numpy_gradients import (
    laplace,
    periodic_laplace,
    fourier_laplace,
    gradient,
    periodic_gradient,
)


class test_gradients(TestCase):
    def test_2d(self):
        N = 128
        grid_size = (N, N)
        L = 1
        dx = 0.2
        input_field = get_2d_sine(grid_size, L)
        arr = scipy.signal.convolve2d(
            input_field, [[0, 1, 0], [1, -4, 1], [0, 1, 0]], boundary="wrap"
        )[1:-1, 1:-1] / (dx**2)
        arr2 = periodic_laplace(input_field, dx)
        assert np.allclose(arr, arr2), f"{arr.shape}, {arr2.shape}"

    def test_3d(self):
        N = 128
        time = 10
        grid_size = (N, N)
        L = 1
        dx = 0.2
        input_field = get_2d_sine(grid_size, L)
        arr = scipy.signal.convolve2d(
            input_field, [[0, 1, 0], [1, -4, 1], [0, 1, 0]], boundary="wrap"
        )[1:-1, 1:-1] / (dx**2)
        input_field = np.stack([input_field] * time, axis=0)
        arr = np.stack([arr] * time, axis=0)
        arr2 = periodic_laplace(input_field, dx)
        assert np.allclose(arr, arr2), f"{arr.shape}, {arr2.shape}"


if __name__ == "__main__":
    pytest.main()



"""
Numerical Properties using NumPy (`numpy_properties`)
=====================================================

This module provides a collection of functions to compute various properties and metrics related to a 2D Hasegawa-Wakatani system. 
It leverages the NumPy library for efficient computations on grid-based data. 
The provided functionalities help in understanding the physical and spectral properties of the system.

Specifically, the module includes:

- **Source** $\\Gamma_n$ as scaler in (x,y) and (kx,ky), as well as array in ky.
- **Sinks** $\\Gamma_c$, $DE$, and $DU$.
- **Energies** including total-, kinetic-, and potential energy.
- **Enstrophy** to quantify the system's vorticity content calculated from Omega and from Phi.
- **Dissipation Metrics** to understand the system's energy dissipation rate over time ($\\partial_t E$, $\\partial_t U$).
- **Spectral Properties** of various metrics for further analysis and verification.

Refer to each function's docstring for detailed information on their use and mathematical formulation.
"""
# Define Frame Properties
# Assume (..., y, x) as shape
import numpy as np
from typing import Tuple
from hw2d.gradients.numpy_gradients import periodic_laplace_N, periodic_gradient


# Gammas


def get_gamma_n(n: np.ndarray, p: np.ndarray, dx: float, dy_p=None) -> float:
    """
    Compute the average particle flux $(\\Gamma_n)$ using the formula:
    $$
        \\Gamma_n = - \\int{\\mathrm{d^2} x \; \\tilde{n} \\frac{\partial \\tilde{\\phi}}{\\partial y}}
    $$

    Args:
        n (np.ndarray): Density (or similar field).
        p (np.ndarray): Potential (or similar field).
        dx (float): Grid spacing.
        dy_p (np.ndarray, optional): Gradient of potential in the y-direction.
            Computed from `p` if not provided.

    Returns:
        float: Computed average particle flux value.
    """
    if dy_p is None:
        dy_p = periodic_gradient(p, dx=dx, axis=-2)  # gradient in y
    gamma_n = -np.mean((n * dy_p), axis=(-1, -2))  # mean over y & x
    return gamma_n


def get_gamma_c(n: np.ndarray, p: np.ndarray, c1: float, dx: float) -> float:
    """
    Compute the sink $\\Gamma_c$ using the formula:
    $$
        \\Gamma_c = c_1 \\int{\\mathrm{d^2} x \; (\\tilde{n} - \\tilde{\\phi})^2}
    $$

    Args:
        n (np.ndarray): Density (or similar field).
        p (np.ndarray): Potential (or similar field).
        c1 (float): Proportionality constant.
        dx (float): Grid spacing.

    Returns:
        float: Computed particle flux value.
    """
    gamma_c = c1 * np.mean((n - p) ** 2, axis=(-1, -2))  # mean over y & x
    return gamma_c


# Spectral Gamma_n


def get_gamma_n_ky(
    n: np.ndarray, p: np.ndarray, dx: float
) -> Tuple[np.ndarray, np.ndarray]:
    """Calculate the spectral components $\\Gamma_n(k_y)$"""
    ky_max = n.shape[-2] // 2
    n_dft = np.fft.fft2(n, norm="ortho", axes=(-2, -1))  # n(ky, kx)
    p_dft = np.fft.fft2(p, norm="ortho", axes=(-2, -1))  # p(ky, kx)
    k_kx, k_ky = np.meshgrid(
        *[np.fft.fftfreq(int(i), d=dx) * 2 * np.pi for i in n.shape[-2:]]
    )  # k(ky, kx)
    gamma_n_k = n_dft * 1j * k_ky * np.conjugate(p_dft)  # gamma_n(ky, kx)
    integrated_gamma_n_k = np.mean(np.real(gamma_n_k), axis=-1)[
        ..., :ky_max
    ]  # gamma_n(ky)
    ky = k_ky[:ky_max, 0]
    return ky, integrated_gamma_n_k


def get_gamma_n_spectrally(n: np.ndarray, p: np.ndarray, dx: float) -> float:
    """Calculate the $\\Gamma_n = \\int{\\mathrm{d} k_y \; \\Gamma_n(k_y)}$"""
    ky, gamma_n_ky = get_gamma_n_ky(n=n, p=p, dx=dx)
    gamma_n = np.mean(gamma_n_ky, axis=-1)  # mean over ky
    return gamma_n


# Energy


def get_energy(n: np.ndarray, phi: np.ndarray, dx: float) -> np.ndarray:
    """Energy of the HW2D system, sum of thermal and kinetic energy
    $$
        E = \\frac{1}{2} \\int{
            \\mathrm {d^2} x \;
            \\left(n^2 + | \\nabla \\phi |^2 \\right)
        }
    $$
    """
    # Squared L2-Norm
    squared_norm_grad_phi = periodic_gradient(phi, dx=dx, axis=-1)**2 + periodic_gradient(
        phi, dx=dx, axis=-2
    )**2
    # Integrate, then divide by 2
    integral = np.mean((n**2) + squared_norm_grad_phi, axis=(-1, -2))
    return integral / 2


# Enstrophy


def get_enstrophy(n: np.ndarray, omega: np.ndarray, dx: float) -> np.ndarray:
    """Enstrophy of the HW2D system
    $$
        \\mathbf{U = \\frac{1}{2} \\int{\\mathrm{d^2} x \; (n - \\Omega)^2}}
                   = \\frac{1}{2} \\int{\\mathrm{d^2} x \; (n^2 - \\nabla^2 \\phi)^2}
    $$
    """
    # omega = omega - np.mean(omega, axis=(-1, -2), keepdims=True)
    integral = np.mean(((n - omega) ** 2), axis=(-1, -2))
    return integral / 2


def get_enstrophy_phi(n: np.ndarray, phi: np.ndarray, dx: float) -> np.ndarray:
    """Enstrophy of the HW2D system from phi
    $$
        \\mathbf{U = \\frac{1}{2} \\int{\\mathrm{d^2} x \; (n^2 - \\nabla^2 \\phi)^2}}
                   = \\frac{1}{2} \\int{\\mathrm{d^2} x \; (n - \\Omega)^2}
    $$
    """
    omega = periodic_laplace_N(phi, dx, N=1)
    omega -= np.mean(omega, axis=(-1, -2), keepdims=True)
    integral = np.mean(((n - omega) ** 2), axis=(-1, -2))
    return integral / 2


def get_D(arr: np.ndarray, nu: float, N: int, dx: float) -> np.ndarray:
    """Calculate the hyperdiffusion coefficient
    $$
        \\nu \; \\nabla F
    $$

    Args:
        arr (np.ndarray): Field to work on
        nu (float): hyperdiffusion coefficient
        N (int): order of hyperdiffusion
        dx (float): grid spacing

    Returns:
        np.ndarray: _description_
    """
    return nu * periodic_laplace_N(arr, dx=dx, N=N)


# Sinks


def get_DE(n: np.ndarray, p: np.ndarray, Dn: np.ndarray, Dp: np.ndarray) -> float:
    """
    $$
        DE = \\int{\\mathrm{d^2} x \; n \; D_n - \\phi \; D_p}
    $$

    Args:
        n (np.ndarray): density field $n$
        p (np.ndarray): potential field $\\phi$
        Dn (np.ndarray): hyperdiffusion of the density field $n$
        Dp (np.ndarray): hyperdiffusion of the potential field $\\phi$

    Returns:
        float: Value of DE
    """
    DE = np.mean(n * Dn - p * Dp, axis=(-1, -2))
    return DE


def get_DU(n: np.ndarray, o: np.ndarray, Dn: np.ndarray, Dp: np.ndarray) -> float:
    """
    $$
        DU = \\int{\\mathrm{d^2} x \; (n - \Omega)  (D_n - D_\\phi)}
    $$

    Args:
        n (np.ndarray): density field $n$
        o (np.ndarray): potential field $\\phi$
        Dn (np.ndarray): hyperdiffusion of the density field $n$
        Dp (np.ndarray): hyperdiffusion of the potential field $\\phi$

    Returns:
        float: Value of D()
    """
    DE = -np.mean((n - o) * (Dn - Dp), axis=(-1, -2))
    return DE


# Time Variation


def get_dE_dt(gamma_n: np.ndarray, gamma_c: np.ndarray, DE: np.ndarray) -> float:
    """
    $$
        \\partial_t E = \\Gamma_n - \\Gamma_c - DE
    $$

    Args:
        gamma_n (np.ndarray): $\\Gamma_n$ 
        gamma_c (np.ndarray): $\\Gamma_c$
        DE (np.ndarray): $DE = n\; D_n - \\phi \; D_p$

    Returns:
        float: Time gradient of energy (in-/outflow)
    """
    return gamma_n - gamma_c - DE


def get_dU_dt(gamma_n: np.ndarray, DU: np.ndarray) -> float:
    """
    $$
        \\partial_t U = \\Gamma_n - DU
    $$

    Args:
        gamma_n (np.ndarray): $\\Gamma_n$
        DU (np.ndarray): $DU$

    Returns:
        float:  Time gradient of enstrophy (in-/outflow)
    """
    return gamma_n - DU


# Spectral Energies


def get_energy_N_ky(n: np.ndarray) -> np.ndarray:
    """thermal energy spectrum
    $$
        E^N(k_y) = \\frac{1}{2} |n(k_y)|^2
    $$
    """
    n_dft = np.fft.fft2(n, norm="ortho")
    # n_dft = np.mean(n_dft, axis=-1)
    E_N_ky = np.abs(n_dft) ** 2 / 2
    E_N_ky = np.mean(E_N_ky, axis=-1)
    return E_N_ky


def get_energy_N_spectrally(n: np.ndarray) -> np.ndarray:
    """thermal energy spectrally integrated
    $$
        E^N = \\int{\\mathrm{d} k_y \; E^N (k_y)} = \\int{\\mathrm{d} k_y \; \\frac{1}{2} |n(k)|^2}
    $$
    """
    E_N_ky = get_energy_N_ky(n)
    E_N = np.mean(E_N_ky, axis=-1)
    return E_N


def get_energy_V_ky(p: np.ndarray, dx: float) -> np.ndarray:
    """kinetic energy spectrum
    $$
        E^V(k_y) = \\frac{1}{2} | k_y \\phi(k_y) |^2
    $$
    """
    k_kx, k_ky = np.meshgrid(
        *[np.fft.fftfreq(int(i), d=dx) * 2 * np.pi for i in p.shape[-2:]]
    )  # k(ky, kx)
    p_dft = np.fft.fft2(p, norm="ortho")
    squared_norm_k = k_ky**2 + k_kx**2
    E_V_ky = np.abs(squared_norm_k * p_dft * np.conjugate(p_dft)) / 2
    E_V_ky = np.mean(E_V_ky, axis=-1)
    return E_V_ky


def get_energy_V_spectrally(p: np.ndarray, dx: float) -> np.ndarray:
    """kinetic energy spectrally integrated
    $$
        E^V = \\int{\\mathrm{d} k_y \; E^V(k_y)} 
            = \\int{\\mathrm{d} k_y \; \\frac{1}{2} |k_y \\phi(k_y) |^2 }
    $$
    """
    E_V_ky = get_energy_V_ky(p, dx=dx)
    E_V = np.mean(E_V_ky, axis=-1)
    return E_V


# Phase Angle Spectra


def get_delta_ky(n: np.ndarray, p: np.ndarray, real=True) -> np.ndarray:
    n_dft = np.fft.fft2(n, norm="ortho")
    p_dft = np.fft.fft2(p, norm="ortho")
    delta_k = np.imag(np.log(np.conjugate(n_dft) * p_dft))
    delta_k = np.mean(delta_k, axis=-1)  # mean in x
    # Get Real component
    delta_k = delta_k[..., : n.shape[1] // 2]
    return delta_k



import numpy as np
import h5py
import os
import pytest

from hw2d.utils.namespaces import Namespace
from hw2d.gradients.numpy_gradients import *
from hw2d.physical_properties.numpy_properties import *
from hw2d.model import HW


def test_energy_conservation_euler(
    dt: float = 0.005, atol: float = 1e-2, rtol: float = 1e-3
):
    # Initial State
    with h5py.File(
        f"{os.path.dirname(os.path.realpath(__file__))}/reference.h5"
    ) as reference:
        plasma = Namespace(
            density=reference["density"][0].astype(np.float64),
            phi=reference["phi"][0].astype(np.float64),
            omega=reference["omega"][0].astype(np.float64),
            age=0,
            dE=0,
            dU=0,
        )
        props = dict(reference.attrs)
    # Make things easier to write
    dx = props["dx"]
    nu = props["nu"]
    N = props["N"]
    # Define Physics
    hw = HW(
        dx=dx,
        N=N,
        nu=nu,
        c1=props["c1"],
        k0=props["k0"],
        poisson_bracket_coeff=props["poisson_bracket_coeff"],
        kappa_coeff=props["kappa_coeff"],
        TEST_CONSERVATION=True,
    )
    plasma["phi"] = hw.get_phi(omega=plasma["omega"], dx=dx)
    # Get initial values
    n = plasma["density"]
    p = plasma["phi"]
    o = plasma["omega"]
    E = get_energy(n=n, phi=p, dx=dx)
    U = get_enstrophy(n=n, omega=o, dx=dx)
    plasma["dE"] = E
    plasma["dU"] = U
    # Simulate a few timesteps to verify
    print(
        f"{'E absolute':>10} ({'E relative':>10})   {'U absolute':>10} ({'U relative':>10})"
    )
    for _ in range(3):
        # Generate Energy Gradients
        gamma_n = get_gamma_n(n=n, p=p, dx=dx)
        gamma_c = get_gamma_c(n=n, p=p, c1=props["c1"], dx=dx)
        Dn = nu * hw.diffuse(n, N=N, dx=dx)
        Dp = nu * hw.diffuse(p, N=N, dx=dx)
        DE = get_DE(n=n, p=p, Dn=Dn, Dp=Dp)
        DU = get_DU(n=n, o=o, Dn=Dn, Dp=Dp)
        dE_dt = get_dE_dt(gamma_n=gamma_n, gamma_c=gamma_c, DE=DE)
        dU_dt = get_dU_dt(gamma_n=gamma_n, DU=DU)
        # Stepping for Plasma State
        plasma = hw.euler_step(plasma, dt=dt, dx=dx)
        # plasma = plasma + dt * hw.gradient_2d(
        #    plasma=plasma, phi=plasma["phi"], dt=0, dx=dx
        # )
        plasma["phi"] = hw.get_phi(omega=plasma["omega"], dx=dx)
        # New
        n = plasma["density"]
        p = plasma["phi"]
        o = plasma["omega"]
        E_exp = plasma.dE
        U_exp = plasma.dU
        E = get_energy(n=n, phi=p, dx=dx)
        U = get_enstrophy(n=n, omega=o, dx=dx)
        print(
            f"{E-E_exp:>10.2e} ({(E-E_exp)/E:>10.2e})   {U-U_exp:>10.2e} ({(U-U_exp)/U:>10.2e})"
        )
        assert (
            np.abs(E_exp - E) < atol
        ), f"AbsoluteError:  E_exp={E_exp:.4e}  E={E:.4e}  {np.abs(E_exp-E):.2e} > {atol}"
        assert (
            np.abs(E_exp - E) / E < rtol
        ), f"RelativeError:  E_exp={E_exp:.4e}  E={E:.4e}  {np.abs(E_exp-E)/E:.2e} > {rtol}"
        assert (
            np.abs(U_exp - U) < atol
        ), f"AbsoluteError:  U_exp={U_exp:.4e}  U={U:.4e}  {np.abs(U_exp-U):.2e} > {atol}"
        assert (
            np.abs(U_exp - U) / U < rtol
        ), f"RelativeError:  U_exp={U_exp:.4e}  U={U:.4e}  {np.abs(U_exp-U)/U:.2e} > {rtol}"


def test_energy_conservation_rk4(
    dt: float = 0.005, atol: float = 1e-2, rtol: float = 1e-3
):
    # Initial State
    with h5py.File(
        f"{os.path.dirname(os.path.realpath(__file__))}/reference.h5"
    ) as reference:
        plasma = Namespace(
            density=reference["density"][0].astype(np.float64),
            phi=reference["phi"][0].astype(np.float64),
            omega=reference["omega"][0].astype(np.float64),
            age=0,
            dE=0,
            dU=0,
        )
        props = dict(reference.attrs)
    # Make things easier to write
    dx = props["dx"]
    nu = props["nu"]
    N = props["N"]
    # Define Physics
    hw = HW(
        dx=dx,
        N=N,
        nu=nu,
        c1=props["c1"],
        k0=props["k0"],
        poisson_bracket_coeff=props["poisson_bracket_coeff"],
        kappa_coeff=props["kappa_coeff"],
        TEST_CONSERVATION=True,
    )
    plasma["phi"] = hw.get_phi(omega=plasma["omega"], dx=dx)
    # Get initial values
    n = plasma["density"]
    p = plasma["phi"]
    o = plasma["omega"]
    E = get_energy(n=n, phi=p, dx=dx)
    U = get_enstrophy(n=n, omega=o, dx=dx)
    plasma["dE"] = E
    plasma["dU"] = U
    # Simulate a few timesteps to verify
    print(
        f"{'E absolute':>10} ({'E relative':>10})   {'U absolute':>10} ({'U relative':>10})"
    )
    for _ in range(3):
        # Generate Energy Gradients
        gamma_n = get_gamma_n(n=n, p=p, dx=dx)
        gamma_c = get_gamma_c(n=n, p=p, c1=props["c1"], dx=dx)
        Dn = nu * hw.diffuse(n, N=N, dx=dx)
        Dp = nu * hw.diffuse(p, N=N, dx=dx)
        DE = get_DE(n=n, p=p, Dn=Dn, Dp=Dp)
        DU = get_DU(n=n, o=o, Dn=Dn, Dp=Dp)
        dE_dt = get_dE_dt(gamma_n=gamma_n, gamma_c=gamma_c, DE=DE)
        dU_dt = get_dU_dt(gamma_n=gamma_n, DU=DU)
        # Stepping for Plasma State
        plasma = hw.rk4_step(plasma, dt=dt, dx=dx)
        plasma["phi"] = hw.get_phi(omega=plasma["omega"], dx=dx)
        # New
        n = plasma["density"]
        p = plasma["phi"]
        o = plasma["omega"]
        E_exp = plasma.dE
        U_exp = plasma.dU
        E = get_energy(n=n, phi=p, dx=dx)
        U = get_enstrophy(n=n, omega=o, dx=dx)
        print(
            f"{E-E_exp:>10.2e} ({(E-E_exp)/E:>10.2e})   {U-U_exp:>10.2e} ({(U-U_exp)/U:>10.2e})"
        )
        assert (
            np.abs(E_exp - E) < atol
        ), f"AbsoluteError:  E_exp={E_exp:.4e}  E={E:.4e}  {np.abs(E_exp-E):.2e} > {atol}"
        assert (
            np.abs(E_exp - E) / E < rtol
        ), f"RelativeError:  E_exp={E_exp:.4e}  E={E:.4e}  {np.abs(E_exp-E)/E:.2e} > {rtol}"
        assert (
            np.abs(U_exp - U) < atol
        ), f"AbsoluteError:  U_exp={U_exp:.4e}  U={U:.4e}  {np.abs(U_exp-U):.2e} > {atol}"
        assert (
            np.abs(U_exp - U) / U < rtol
        ), f"RelativeError:  U_exp={U_exp:.4e}  U={U:.4e}  {np.abs(U_exp-U)/U:.2e} > {rtol}"


if __name__ == "__main__":
    test_energy_conservation_euler()
    test_energy_conservation_rk4()



"""
`hw2d.initializations`: Initialization methods for 2D patterns in the HW2D system.

This module provides a collection of functions to generate initial patterns or noise
for the HW2D system. The methods encompass both deterministic patterns (like sine waves)
and stochastic patterns (like Fourier noise). These initializations can be useful for
starting simulations, testing, and visualization in the context of the HW2D system.

Functions:
    - `get_2d_sine`: Generates a 2D sine wave pattern over a specified grid.
    - `get_fft_noise`: Creates a 2D noise pattern using the Fast Fourier Transform.

Example Usage:
    >>> # Generate a 2D sine pattern
    >>> sine_pattern = get_2d_sine((256, 256), 1.0)
    
    >>> # Generate a 2D noise pattern using FFT
    >>> noise_pattern = get_fft_noise((256, 256), 1.0, 1.0, min_frequency=0.1, max_frequency=2.0)
"""



import numpy as np
import numba

precision = np.float64
complex_precision = np.complex128


def fourier_poisson_double(tensor: np.ndarray, dx: float, times: int = 1) -> np.ndarray:
    """Inverse operation to `fourier_laplace`."""
    tensor = np.array(tensor, dtype=complex_precision)
    frequencies = np.fft.fft2(tensor)

    k = fftfreq_sq(tensor.shape)
    result_comp = core_computation(frequencies, k, dx, times)

    result = np.real(np.fft.ifft2(result_comp))
    return result.astype(precision)


@numba.jit(nopython=True)
def core_computation(
    frequencies: np.ndarray, k: np.ndarray, dx: float, times: int = 1
) -> np.ndarray:
    fft_laplace = -((2 * np.pi) ** 2) * k
    # Avoiding the use of np.inf for now. Set to a very large number.
    fft_laplace[0, 0] = 1e14

    frequencies = np.where(fft_laplace == 0, 0, frequencies)
    result = frequencies / (fft_laplace**times)

    return result * dx**2


@numba.jit(nopython=True)
def fftfreq_sq(resolution: np.ndarray) -> np.ndarray:
    dim_x, dim_y = resolution
    freq_x = custom_fftfreq(dim_x)
    freq_y = custom_fftfreq(dim_y)

    # Equivalent to:  k_sq = freq_x**2 + freq_y**2
    k_sq = np.empty((dim_x, dim_y), dtype=precision)
    for i in numba.prange(dim_x):
        for j in numba.prange(dim_y):
            k_sq[i, j] = freq_x[i] ** 2 + freq_y[j] ** 2

    return k_sq


@numba.jit(nopython=True)
def custom_fftfreq(n: np.ndarray) -> np.ndarray:
    """Custom FFT frequency function to replicate np.fft.fftfreq for Numba."""
    results = np.empty(n, dtype=np.int16)
    N = (n - 1) // 2 + 1
    results[:N] = np.arange(0, N)
    results[N:] = np.arange(-(n // 2), 0)
    return results / n



import numpy as np
import pytest
from hw2d.gradients import numpy_gradients
from hw2d.gradients.numba_gradients import (
    laplace,
    periodic_laplace,
    laplace_N,
    periodic_laplace_N,
    gradient,
    periodic_gradient,
)


TEST_SIZE = 32


def test_laplace_nb():
    a = np.random.rand(TEST_SIZE, TEST_SIZE)
    dx = float(np.random.rand(1))
    result = laplace(np.pad(a, 1), dx)
    expected = numpy_gradients.laplace(np.pad(a, 1), dx)
    assert np.array_equal(result, expected)


def test_periodic_laplace():
    a = np.random.rand(TEST_SIZE, TEST_SIZE)
    dx = float(np.random.rand(1))
    result = periodic_laplace(a, dx)
    expected = numpy_gradients.periodic_laplace(a, dx)
    assert np.array_equal(result, expected)


def test_periodic_laplace_2():
    a = np.random.rand(TEST_SIZE, TEST_SIZE)
    dx = float(np.random.rand(1))
    result = periodic_laplace_N(a, N=2, dx=dx)
    expected = numpy_gradients.periodic_laplace_N(a, dx, N=2)
    assert np.array_equal(result, expected)


def test_periodic_laplace_3():
    a = np.random.rand(TEST_SIZE, TEST_SIZE)
    dx = float(np.random.rand(1))
    result = periodic_laplace_N(a, N=3, dx=dx)
    expected = numpy_gradients.periodic_laplace_N(a, dx, N=3)
    assert np.array_equal(result, expected)


def test_gradient():
    a = np.random.rand(TEST_SIZE, TEST_SIZE)
    dx = float(np.random.rand(1))
    result = gradient(a, dx, axis=0)
    expected = numpy_gradients.gradient(a, dx, axis=0)
    assert np.array_equal(result, expected)


def test_periodic_gradient():
    a = np.random.rand(TEST_SIZE, TEST_SIZE)
    dx = float(np.random.rand(1))
    result = periodic_gradient(a, dx, axis=0)
    expected = numpy_gradients.periodic_gradient(a, dx, axis=0)
    assert np.array_equal(result, expected)


if __name__ == "__main__":
    pytest.main()



from typing import Iterable, Dict, Tuple, Any, Callable
import h5py
import fire
import numpy as np
import matplotlib.pyplot as plt
from tqdm import tqdm
from matplotlib.colors import ListedColormap
from matplotlib import animation
from mpl_toolkits.axes_grid1 import make_axes_locatable


def get_extended_viridis(vals: int = 600) -> ListedColormap:
    """
    Generate an extended viridis colormap.

    Args:
        vals (int): Number of color values. Default is 600.

    Returns:
        ListedColormap: A colormap object.
    """
    VIRIDIS_EXTENDED = [
        [0.0, 255, 200, 100],
        [0.13, 255, 153, 51],
        [0.25, 230, 5, 40],
        [0.38, 150, 3, 62],
        [0.5, 68, 1, 84],
        [0.55, 72, 33, 115],
        [0.59, 67, 62, 133],
        [0.64, 56, 88, 140],
        [0.68, 45, 112, 142],
        [0.73, 37, 133, 142],
        [0.77, 30, 155, 138],
        [0.82, 42, 176, 127],
        [0.86, 82, 197, 105],
        [0.90, 34, 213, 73],
        [0.95, 194, 223, 35],
        [1.0, 253, 231, 37],
    ]
    VE = np.array(VIRIDIS_EXTENDED)
    VE[:, 1:] = VE[:, 1:] / 256
    ve_list = []
    for i in range(1, len(VIRIDIS_EXTENDED)):
        pts = int((VE[i, 0] - VE[i - 1, 0]) * vals)
        ve_list.append(
            np.stack(
                [
                    np.linspace(VE[i - 1, 1], VE[i, 1], pts, endpoint=False),
                    np.linspace(VE[i - 1, 2], VE[i, 2], pts, endpoint=False),
                    np.linspace(VE[i - 1, 3], VE[i, 3], pts, endpoint=False),
                ],
                axis=-1,
            )
        )
    VE = np.vstack(ve_list)
    return ListedColormap(VE)


def new_cbar_max(nm: float, pm: float) -> float:
    """
    Adjust maximum of colorbar using sqrt diff scaling.

    Args:
        nm (float): Negative maximum.
        pm (float): Positive maximum.

    Returns:
        float: Adjusted maximum.
    """
    if pm > nm:
        nm = nm + np.sqrt(pm - nm)
    return nm


def new_cbar_max_smooth(nm: float, pm: float) -> float:
    """
    Adjust maximum of colorbar using average scaling for a smoother effect.

    Args:
        nm (float): Negative maximum.
        pm (float): Positive maximum.

    Returns:
        float: Adjusted maximum.
    """
    if pm > nm:
        nm = (nm + pm) / 2
    return nm


def setup_visualization(
    axarr: Iterable[Any],
    plasma_steps: Iterable[Dict[str, float]],
    params: Dict[str, float],
    plot_order: Iterable[str],
    cmap: Any,
) -> Tuple[Iterable[Any], Iterable[Any], Iterable[Any], Iterable[Any]]:
    """Setup initial chart objects."""
    ims, cxs, cbs, txs = [], [], [], []
    for i, ax in enumerate(axarr):
        ax.xaxis.set_tick_params(
            which="both", bottom=False, top=False, labelbottom=False
        )
        ax.yaxis.set_tick_params(
            which="both", bottom=False, top=False, labelbottom=False
        )
        cxs.append(make_axes_locatable(ax).append_axes("right", "5%", "1%"))
        ims.append(ax.imshow(plasma_steps[plot_order[i]][0], origin="lower", cmap=cmap))
        cbs.append(plt.colorbar(ims[i], cax=cxs[i]))
        txs.append(ax.set_title(f"{plot_order[i]}"))
    return ims, cxs, cbs, txs


def setup_figure(title: str) -> Tuple[plt.Figure, np.ndarray]:
    """
    Setup the figure for visualization.

    Args:
        title (str): Title for the figure.

    Returns:
        Tuple containing the figure and array of axes.
    """
    fig, axarr = plt.subplots(1, 3, figsize=(12, 4.35), sharex=True, sharey=True)
    fig.suptitle(title)
    return fig, axarr


def time_to_length(time: int, dt: float) -> int:
    """Convert time to length based on the given time step."""
    return int(round(time / dt))


def create_movie(
    input_filename: str,
    output_filename: str,
    t0: int = 0,
    t1: int = None,
    zero_omega: bool = False,
    plot_order: Iterable[str] = ["density", "omega", "phi"],
    cmap: ListedColormap = get_extended_viridis(),
    min_fps: float = 10,
    dpi: int = 75,
    speed: float = 1,
) -> None:
    # Setup details
    plasma_steps = h5py.File(input_filename, "r")
    params = dict(plasma_steps.attrs)
    title = generate_title(params)
    if t1 is None:
        t1 = (len(plasma_steps[plot_order[0]]) - 1) * params["frame_dt"]
    t0_idx = time_to_length(t0 - params.get("t0", 0), params["frame_dt"])
    t1_idx = time_to_length(t1 - params.get("t0", 0), params["frame_dt"])
    # Determine fps and step size to use
    fps = int(speed / params["frame_dt"])
    print(f"{speed}t/s  @  {params['frame_dt']} t/frame  implies {fps} frames/s")
    frame_steps = max(1, fps // min_fps)
    fps /= frame_steps
    print(
        f"with min_fps {min_fps} we take every {frame_steps} frames still generate {fps} frame/s"
    )
    # Define Progress
    total = int((t1_idx - t0_idx) // frame_steps)
    frame_range = range(t0_idx, t1_idx + 1, frame_steps)
    pbar = tqdm(total=total, smoothing=0.3)
    # Setup Figure & Animation
    fig, axarr = setup_figure(title)
    ims, cxs, cbs, txs = setup_visualization(
        axarr, plasma_steps, params, plot_order, cmap
    )
    fig.subplots_adjust(top=0.94, bottom=0, right=0.95, left=0.01, hspace=0, wspace=0.2)
    # Animation
    writer = animation.writers["ffmpeg"](fps=fps, metadata=dict(artist="Robin Greif"))

    def animate(t_idx: int):
        """Update data for animation."""
        for i, ax in enumerate(axarr):
            field_name = plot_order[i]
            arr = plasma_steps[field_name][t_idx]
            if zero_omega and field_name == "omega":
                arr -= np.mean(arr)
            vmax = np.max(arr)
            vmin = np.min(arr)
            pmin, pmax = ims[i].get_clim()
            pm = np.max([np.abs(pmin), pmax])
            nm = np.max([np.abs(vmin), vmax])
            nm = new_cbar_max_smooth(nm, pm)
            ims[i].set_data(arr)
            ims[i].set_clim(-nm, nm)
            txs[i].set_text(f"{field_name} (t={t_idx*params['frame_dt']:.0f})")
        pbar.update(1)

    ani = animation.FuncAnimation(fig, animate, frames=frame_range)
    # Save Movie
    output_filename = f"{output_filename}_dpi={dpi}_fps={fps:.0f}_speed={speed:.0f}_t0={t0}_t1={t1}.mp4"
    ani.save(output_filename, writer=writer, dpi=dpi)
    plt.close()
    plasma_steps.close()
    print(f"{output_filename}")


def generate_title(params: Dict[str, float]) -> str:
    return ",  ".join(
        [
            f"c1={params['c1']}",
            # f"L={params['L']:.2f}",
            f"k0={params['k0']:.2f}",
            f"pts={params['grid_pts']:.0f}",
            f"dt={params['dt']}",
            f"N={params['N']:.0f}",
            f"nu={params['nu']:.2g}",
        ]
    )


def main(
    input_path: str,
    output_path: str,
    t0: int = 0,
    t1: int or None = None,
    plot_order: Iterable = ("density", "omega", "phi"),
    dpi: int = 75,
    min_fps: int = 10,
    speed: int = 10,
) -> None:
    create_movie(
        input_filename=input_path,
        output_filename=output_path,
        t0=t0,
        t1=t1,
        plot_order=plot_order,
        min_fps=min_fps,
        dpi=dpi,
        speed=speed,
    )


if __name__ == "__main__":
    fire.Fire(main)



import numpy as np
from numba import stencil, jit, prange


@stencil
def jpp(zeta: np.ndarray, psi: np.ndarray, d: float) -> np.ndarray:
    return (
        (zeta[0, 1] - zeta[0, -1]) * (psi[1, 0] - psi[-1, 0])
        - (zeta[1, 0] - zeta[-1, 0]) * (psi[0, 1] - psi[0, -1])
    ) / (4 * d**2)


@stencil
def jpx(zeta: np.ndarray, psi: np.ndarray, d: float) -> np.ndarray:
    return (
        zeta[0, 1] * (psi[1, 1] - psi[-1, 1])
        - zeta[0, -1] * (psi[1, -1] - psi[-1, -1])
        - zeta[1, 0] * (psi[1, 1] - psi[1, -1])
        + zeta[-1, 0] * (psi[-1, 1] - psi[-1, -1])
    ) / (4 * d**2)


@stencil
def jxp(zeta: np.ndarray, psi: np.ndarray, d: float) -> np.ndarray:
    return (
        zeta[1, 1] * (psi[1, 0] - psi[0, 1])
        - zeta[-1, -1] * (psi[0, -1] - psi[-1, 0])
        - zeta[1, -1] * (psi[1, 0] - psi[0, -1])
        + zeta[-1, 1] * (psi[0, 1] - psi[-1, 0])
    ) / (4 * d**2)


@jit(nopython=True, parallel=True, nogil=True)
def arakawa(zeta: np.ndarray, psi: np.ndarray, dx: float) -> np.ndarray:
    return (jpp(zeta, psi, dx) + jpx(zeta, psi, dx) + jxp(zeta, psi, dx)) / 3


def periodic_arakawa(zeta, psi, dx):
    return arakawa(np.pad(zeta, 1, mode="wrap"), np.pad(psi, 1, mode="wrap"), dx)[
        1:-1, 1:-1
    ]


# Full Stencil
@stencil
def arakawa_stencil(zeta: np.ndarray, psi: np.ndarray) -> np.ndarray:
    return (
        zeta[0, 1] * (psi[1, 0] - psi[-1, 0] + psi[1, 1] - psi[-1, 1])
        - zeta[0, -1] * (psi[1, 0] - psi[-1, 0] + psi[1, -1] - psi[-1, -1])
        - zeta[1, 0] * (psi[0, 1] - psi[0, -1] + psi[1, 1] - psi[1, -1])
        + zeta[-1, 0] * (psi[0, 1] - psi[0, -1] + psi[-1, 1] - psi[-1, -1])
        + zeta[-1, 1] * (psi[0, 1] - psi[-1, 0])
        + zeta[1, 1] * (psi[1, 0] - psi[0, 1])
        - zeta[1, -1] * (psi[1, 0] - psi[0, -1])
        - zeta[-1, -1] * (psi[0, -1] - psi[-1, 0])
    )


@jit(nopython=True)
def arakawa_stencil_full(zeta: np.ndarray, psi: np.ndarray, dx: float) -> np.ndarray:
    return (arakawa_stencil(zeta, psi))[1:-1, 1:-1] / (12 * dx**2)


def periodic_arakawa_stencil(
    zeta: np.ndarray, psi: np.ndarray, dx: float
) -> np.ndarray:
    return arakawa_stencil_full(
        np.pad(zeta, 1, mode="wrap"), np.pad(psi, 1, mode="wrap"), dx
    )


## Vectorized


@jit(nopython=True)
def arakawa_vec(zeta: np.ndarray, psi: np.ndarray, dx: float) -> np.ndarray:
    """2D periodic first-order Arakawa
    requires 1 cell padded input on each border"""
    return (
        zeta[1:-1, 2:] * (psi[2:,1:-1] - psi[0:-2, 1:-1] + psi[2:, 2:] - psi[0:-2, 2:])
        - zeta[1:-1,0:-2]
        * (psi[2:, 1:-1] - psi[0:-2, 1:-1] + psi[2:, 0:-2] - psi[0:-2, 0:-2])
        - zeta[2:, 1:-1]
        * (psi[1:-1, 2:] - psi[1:-1, 0:-2] + psi[2:, 2:] - psi[2:, 0:-2])
        + zeta[0:-2, 1:-1]
        * (psi[1:-1, 2:] - psi[1:-1, 0:-2] + psi[0:-2, 2:] - psi[0:-2, 0:-2])
        + zeta[0:-2, 2:] * (psi[1:-1, 2:] - psi[0:-2, 1:-1])
        + zeta[2:, 2:] * (psi[2:, 1:-1] - psi[1:-1, 2:])
        - zeta[2:, 0:-2] * (psi[2:, 1:-1] - psi[1:-1, 0:-2])
        - zeta[0:-2, 0:-2] * (psi[1:-1, 0:-2] - psi[0:-2, 1:-1])
    ) / (12 * dx**2)


def periodic_arakawa_vec(zeta: np.ndarray, psi: np.ndarray, dx: float) -> np.ndarray:
    return arakawa_vec(np.pad(zeta, 1, mode="wrap"), np.pad(psi, 1, mode="wrap"), dx)



"""
Fourier-based Poisson Solvers with NumPy
========================================

This module provides a set of functions to solve the Poisson equation using the Fourier transform approach and the NumPy library. The Fourier method is especially beneficial for periodic domains and spectral methods, as it can retrieve the original field from its gradient or Laplace efficiently in the spectral space.

Main functionalities include:

- `fourier_poisson_double`: Solves the Poisson equation using double precision (complex128).
- `fourier_poisson_single`: Solves the Poisson equation using single precision (complex64).
- `fourier_poisson_numpy`: A more general Poisson solver that auto-detects the input type.

These functions are designed for both prototyping and production-level simulations, offering a balance between accuracy and performance. They are particularly well-suited for large-scale simulations in periodic domains.
"""

import numpy as np


def fourier_poisson_double(tensor: np.ndarray, dx: float, times: int = 1) -> np.ndarray:
    """Inverse operation to `fourier_laplace`."""
    tensor = np.array(tensor, dtype=np.complex128)
    frequencies = np.fft.fft2(tensor)

    k = np.meshgrid(*[np.fft.fftfreq(int(n)) for n in tensor.shape], indexing="ij")
    k = np.stack(k, -1)
    k = np.sum(k**2, axis=-1)
    fft_laplace = -((2 * np.pi) ** 2) * k
    fft_laplace[0, 0] = np.inf

    with np.errstate(divide="ignore", invalid="ignore"):
        result = frequencies / (fft_laplace**times)
    result = np.where(fft_laplace == 0, 0, result)

    result = np.real(np.fft.ifft2(result))
    return (result * dx**2).astype(np.float64)


def fourier_poisson_single(tensor: np.ndarray, dx: float, times: int = 1) -> np.ndarray:
    """Inverse operation to `fourier_laplace`."""
    tensor = np.array(tensor, dtype=np.complex64)
    frequencies = np.fft.fft2(tensor)

    k = np.meshgrid(*[np.fft.fftfreq(int(n)) for n in tensor.shape], indexing="ij")
    k = np.stack(k, -1)
    k = np.sum(k**2, axis=-1)
    fft_laplace = -((2 * np.pi) ** 2) * k
    fft_laplace[0, 0] = np.inf

    with np.errstate(divide="ignore", invalid="ignore"):
        result = frequencies / (fft_laplace**times)
    result = np.where(fft_laplace == 0, 0, result)

    result = np.real(np.fft.ifft2(result))
    return (result * dx**2).astype(np.float32)


def fourier_poisson_numpy(grid: np.ndarray, dx: float, times: int = 1) -> np.ndarray:
    """
    Inverse operation to `fourier_laplace`.

    Args:
      grid: numpy array
      dx: float or list or tuple
      times: int (Default value = 1)

    Returns:
      result: numpy array
    """
    # Convert grid to complex type
    grid_complex = np.asarray(grid, dtype=np.complex128)

    # Compute the FFT of the grid
    frequencies = np.fft.fftn(grid_complex)

    # Compute squared frequency magnitude
    k_squared = np.sum(
        np.array(np.meshgrid(*[np.fft.fftfreq(dim) for dim in grid.shape])) ** 2, axis=0
    )

    # Compute inverse of squared frequency multiplied by -(2 * np.pi)^2
    fft_laplace = -((2 * np.pi) ** 2) * k_squared

    # Handle division by zero safely and compute the inverse FFT
    divisor = fft_laplace**times
    with np.errstate(divide="ignore", invalid="ignore"):
        safe_division = np.where(divisor != 0, frequencies / divisor, 0)
    result = np.real(np.fft.ifftn(safe_division))

    # Multiply by dx squared
    result *= np.prod(dx) ** 2

    # Cast result to the original datatype of grid
    return result.astype(grid.dtype)



import numpy as np
import h5py
import fire
from tqdm import tqdm
from typing import List
from functools import partial

from hw2d.physical_properties.numpy_properties import (
    get_gamma_n,
    get_gamma_n_spectrally,
    get_gamma_c,
    get_energy,
    get_energy_N_spectrally,
    get_energy_V_spectrally,
    get_enstrophy,
    get_enstrophy_phi,
)


def add_data(hf, i, batch_size, name, data, debug, selection):
    if name in selection:
        data = data()
        hf[name][i : i + batch_size] = data
        if debug:
            print(f"{name}: {np.mean(data):.2e}", end="  |  ")


def calculate_properties(
    file_path: str = "",
    batch_size: int = 100,
    property_list: List = [
        "gamma_n",
        "gamma_n_spectral",
        "gamma_c",
        "energy",
        "thermal_energy",
        "kinetic_energy",
        "enstrophy",
        "enstrophy_phi",
    ],
    force_recompute: bool = True,
    is_debug: bool = False,
):
    with h5py.File(file_path, "r+") as h5_file:
        # Parameters
        parameters = dict(h5_file.attrs)
        L = 2 * np.pi / parameters["k0"]
        dx = L / parameters["x_save"]
        c1 = parameters["c1"]
        dt = parameters["dt"]
        steps = len(h5_file["density"])
        # Create Properties in Dataset
        selection = []
        for property_name in property_list:
            if property_name in h5_file.keys():
                print(f"Dataset exists:  {property_name}")
                if force_recompute:
                    del h5_file[property_name]
            if property_name not in h5_file.keys():
                selection.append(property_name)
                h5_file.create_dataset(property_name, (steps,))
                print(f"Created Dataset:  {property_name}")
        # Set up iterator
        iterator = range(0, steps, batch_size)
        if not is_debug:
            iterator = tqdm(iterator)
        # Run Through in Batches
        for i in iterator:
            add = partial(
                add_data,
                hf=h5_file,
                i=i,
                batch_size=batch_size,
                debug=is_debug,
                selection=selection,
            )
            n = h5_file["density"][i : i + batch_size]
            p = h5_file["phi"][i : i + batch_size]
            o = h5_file["omega"][i : i + batch_size]
            if is_debug:
                print(
                    f"Timeframe ({i*dt}-{(i + batch_size)*dt})",
                    end="\n  ",
                )
            # Add properties if they were selected
            add(name="gamma_n", data=partial(get_gamma_n, n=n, p=p, dx=dx))
            add(
                name="gamma_n_spectral",
                data=partial(get_gamma_n_spectrally, n=n, p=p, dx=dx),
            )
            add(name="gamma_c", data=partial(get_gamma_c, n=n, p=p, c1=c1, dx=dx))
            add(name="energy", data=partial(get_energy, n=n, phi=p, dx=dx))
            add(name="thermal_energy", data=partial(get_energy_N_spectrally, n=n))
            add(
                name="kinetic_energy", data=partial(get_energy_V_spectrally, p=p, dx=dx)
            )
            add(name="enstrophy", data=partial(get_enstrophy, n=n, omega=o, dx=dx))
            add(
                name="enstrophy_phi", data=partial(get_enstrophy_phi, n=n, phi=p, dx=dx)
            )
            if is_debug:
                print()

    return


if __name__ == "__main__":
    fire.Fire(calculate_properties)



"""
NumPy-based Arakawa Scheme Implementations
==========================================

This module provides implementations of the Arakawa scheme for computing the Poisson bracket using the NumPy library. 
The functions are tailored for a 2D domain and utilize efficient NumPy array operations to achieve accurate and fast computations.

The module includes:

- Basic Jacobian functions with different discretization strategies (`jpp`, `jpx`, `jxp`).
- A main function (`arakawa`) which computes the Poisson bracket as an average of the Jacobians.
- Periodically padded implementations (`periodic_arakawa`).
- A vectorized implementation for improved performance (`arakawa_vec`).

Note: The functions in this module require the vorticity field (`zeta`) and the stream function field (`psi`) as primary inputs.
"""

import numpy as np


def jpp(zeta, psi, dx, i, j):
    """
    Compute the Jacobian using centered differences for both fields.

    Args:
        zeta (np.ndarray): Vorticity field.
        psi (np.ndarray): Stream function field.
        dx (float): Grid spacing.
        i, j (int): Indices for spatial location (x, y respectively)

    Returns:
        float: Jacobian value at (i, j).
    """
    return (
        (zeta[j, i + 1] - zeta[j,i - 1]) * (psi[j + 1, i] - psi[j - 1, i])
        - (zeta[j + 1, i] - zeta[j - 1, i]) * (psi[j, i + 1] - psi[j, i - 1])
    ) / (4 * dx**2)


def jpx(zeta, psi, dx, i, j):
    """
    Compute the Jacobian using centered differences for the vorticity field
    and staggered differences for the stream function field.

    Args:
        zeta (np.ndarray): Vorticity field.
        psi (np.ndarray): Stream function field.
        dx (float): Grid spacing.
        i, j (int): Indices for spatial location (x, y respectively)

    Returns:
        float: Jacobian value at (i, j).
    """
    return (
        zeta[j, i + 1] * (psi[j + 1, i + 1] - psi[j - 1, i + 1])
        - zeta[j, i - 1] * (psi[j + 1, i - 1] - psi[j - 1, i - 1])
        - zeta[j + 1, i] * (psi[j + 1, i + 1] - psi[j + 1, i - 1])
        + zeta[j - 1, i] * (psi[j - 1, i + 1] - psi[j - 1, i - 1])
    ) / (4 * dx**2)


def jxp(zeta, psi, dx, i, j):
    """
    Compute the Jacobian using staggered differences for the vorticity field
    and centered differences for the stream function field.

    Args:
        zeta (np.ndarray): Vorticity field.
        psi (np.ndarray): Stream function field.
        dx (float): Grid spacing.
        i, j (int): Indices for spatial location (x, y respectively)

    Returns:
        float: Jacobian value at (i, j).
    """
    return (
        zeta[j + 1, i + 1] * (psi[j + 1, i] - psi[j, i + 1])
        - zeta[j - 1, i - 1] * (psi[j, i - 1] - psi[j - 1, i])
        - zeta[j + 1, i - 1] * (psi[j + 1, i] - psi[j, i - 1])
        + zeta[j - 1, i + 1] * (psi[j, i + 1] - psi[j - 1, i])
    ) / (4 * dx**2)


def arakawa(zeta, psi, dx):
    """
    Compute the Poisson bracket as an average Jacobian using the Arakawa scheme.

    Args:
        zeta (np.ndarray): Vorticity field.
        psi (np.ndarray): Stream function field.
        dx (float): Grid spacing.

    Returns:
        np.ndarray: Compute the Poisson bracket as an average Jacobian over the entire grid.
    """
    val = np.empty_like(zeta)
    for j in range(1, zeta.shape[0] - 1):  # j = y
        for i in range(1, zeta.shape[1] - 1):  # i = x
            val[j][i] = (
                jpp(zeta, psi, dx, i, j)
                + jpx(zeta, psi, dx, i, j)
                + jxp(zeta, psi, dx, i, j)
            )
    return val / 3


def periodic_arakawa(zeta, psi, dx):
    """
    Compute the Arakawa Scheme with periodic boundary conditions.

    Args:
        zeta (np.ndarray): Vorticity field.
        psi (np.ndarray): Stream function field.
        dx (float): Grid spacing.

    Returns:
        np.ndarray: Compute the Poisson bracket as an average Jacobian over the grid without padding.
    """
    return arakawa(np.pad(zeta, 1, mode="wrap"), np.pad(psi, 1, mode="wrap"), dx)[
        1:-1, 1:-1
    ]


## Vectorized


def arakawa_vec(zeta, psi, dx):
    """
    Compute the Poisson bracket (Jacobian) of vorticity and stream function
    using a vectorized version of the Arakawa scheme. This function is designed
    for a 2D periodic domain and requires a 1-cell padded input on each border.

    Args:
        zeta (np.ndarray): Vorticity field with padding.
        psi (np.ndarray): Stream function field with padding.
        dx (float): Grid spacing.

    Returns:
        np.ndarray: Discretized Poisson bracket (Jacobian) over the grid.
    """
    return (
        zeta[1:-1, 2:] * (psi[2:,1:-1] - psi[0:-2, 1:-1] + psi[2:, 2:] - psi[0:-2, 2:])
        - zeta[1:-1,0:-2]
        * (psi[2:, 1:-1] - psi[0:-2, 1:-1] + psi[2:, 0:-2] - psi[0:-2, 0:-2])
        - zeta[2:, 1:-1]
        * (psi[1:-1, 2:] - psi[1:-1, 0:-2] + psi[2:, 2:] - psi[2:, 0:-2])
        + zeta[0:-2, 1:-1]
        * (psi[1:-1, 2:] - psi[1:-1, 0:-2] + psi[0:-2, 2:] - psi[0:-2, 0:-2])
        + zeta[0:-2, 2:] * (psi[1:-1, 2:] - psi[0:-2, 1:-1])
        + zeta[2:, 2:] * (psi[2:, 1:-1] - psi[1:-1, 2:])
        - zeta[2:, 0:-2] * (psi[2:, 1:-1] - psi[1:-1, 0:-2])
        - zeta[0:-2, 0:-2] * (psi[1:-1, 0:-2] - psi[0:-2, 1:-1])
    ) / (12 * dx**2)


def periodic_arakawa_vec(zeta, psi, dx):
    """
    Compute the Poisson bracket (Jacobian) of vorticity and stream function for a 2D periodic
    domain using a vectorized version of the Arakawa scheme. This function automatically
    handles the required padding.

    Args:
        zeta (np.ndarray): Vorticity field.
        psi (np.ndarray): Stream function field.
        dx (float): Grid spacing.

    Returns:
        np.ndarray: Discretized Poisson bracket (Jacobian) over the grid without padding.
    """
    return arakawa_vec(np.pad(zeta, 1, mode="wrap"), np.pad(psi, 1, mode="wrap"), dx)



import numpy as np
import pytest
from hw2d.initializations.fourier_noise import get_fft_noise
from hw2d.poisson_solvers.numpy_fourier_poisson import fourier_poisson_double, fourier_poisson_numpy
fourier_poisson_np = fourier_poisson_double
from hw2d.poisson_solvers.numba_fourier_poisson import fourier_poisson_double
fourier_poisson_nb = fourier_poisson_double
from hw2d.gradients.numpy_gradients import periodic_laplace


TEST_SIZE = 32
defs = dict(atol=1e-14, rtol=1e-20, equal_nan=False)
defs = dict(atol=1e-7, rtol=1e-20, equal_nan=False)


def test_fourier_poisson():
    a = np.random.rand(TEST_SIZE,TEST_SIZE)
    dx = float(np.random.rand(1))
    result = fourier_poisson_nb(a, dx)
    expected = fourier_poisson_np(a, dx)
    print(result)
    print(expected)
    # Assert that shapes match
    assert result.shape == expected.shape, f"Expected shape: {expected.shape}, but got {result.shape}"

    # Assert that values are close
    is_close = np.isclose(result, expected, **defs)

    # If not all values are close, print detailed info
    if not np.all(is_close):
        abs_diff = np.abs(result - expected)
        rel_diff = abs_diff / np.abs(expected)
        max_abs_diff = np.max(abs_diff)
        max_rel_diff = np.max(rel_diff)
        
        msg = (f"Max absolute difference: {max_abs_diff}\n"
               f"Max relative difference: {max_rel_diff}\n")
        
        assert False, msg


def test_poisson():
    def get_2d_sine(grid_size, L):
        indices = np.array(np.meshgrid(*list(map(range, grid_size))))
        phys_coord = indices.T*L/(grid_size[0])  # between [0, L)
        x, y = phys_coord.T
        d = np.sin(x+1) * np.sin(y+1)
        return d

    N = 512
    L = 2*np.pi
    dx = L/N

    sin = get_2d_sine((N,N), L)
    laplace_sin = periodic_laplace(sin, dx)
    pred_sins = {
        "fourier_poisson_np": fourier_poisson_np(laplace_sin, dx),
        "fourier_poisson_nb": fourier_poisson_nb(laplace_sin, dx),
        "fourier_poisson_numpy": fourier_poisson_numpy(laplace_sin, dx)
    }

    for k, v in pred_sins.items():
        print(np.max(np.abs(v-sin)), np.mean(v-sin))

    fft_noise = get_fft_noise(
        (N, N),
        L,
        1,
        min_wavelength=dx * 12,
        max_wavelength=dx * N * 100,
        factor=2,
    )
    laplace_fft_noise = periodic_laplace(fft_noise, dx)
    pred_fft_noise = {
        "fourier_poisson_np": fourier_poisson_np(laplace_fft_noise, dx),
        "fourier_poisson_nb": fourier_poisson_nb(laplace_fft_noise, dx),
        "fourier_poisson_numpy":  fourier_poisson_numpy(laplace_fft_noise, dx)
    }

    for k, v in pred_fft_noise.items():
        print(np.max(np.abs(v-fft_noise)), np.mean(v-fft_noise))


if __name__ == '__main__':
    pytest.main()



from typing import Dict, Tuple, List, Union
import numpy as np
import matplotlib.pyplot as plt
from tqdm import tqdm
import fire
import h5py

from hw2d.utils.plot.movie import get_extended_viridis
from hw2d.utils.latex_format import latex_format


def is_zero_included(vals: np.ndarray, frac: float = 0.025) -> None:
    min_yvals = np.min(vals)
    max_yvals = np.max(vals)
    range_vals = max_yvals - min_yvals
    if (min_yvals - frac * range_vals <= 0) and (0 <= max_yvals - frac * range_vals):
        return True
    else:
        return False


def add_axes(
    x_vals: np.ndarray,
    y_vals: np.ndarray,
    ax: plt.Axes,
    linewidth: float = 1,
    linestyle: str = "solid",
    alpha: float = 1,
    color: str = "black",
) -> None:
    if is_zero_included(x_vals):
        ax.axvline(
            0, linewidth=linewidth, linestyle=linestyle, alpha=alpha, color=color
        )
    if is_zero_included(y_vals):
        ax.axhline(
            0, linewidth=linewidth, linestyle=linestyle, alpha=alpha, color=color
        )


def plot_timeline(values: np.ndarray, t0: float, dt: float, ax: plt.Axes, **kwargs):
    """plot values over time with proper axis adjustments

    :param values: y-values
    :type values: array-like
    :param t0: first time (only used to adjust x-axis)
    :type t0: float
    :param dt: time step (only used to adjust x-axis)
    :type dt: float
    :param ax: axis to plot on
    :type ax: matplotlib.axis
    :return: axis.plot return
    :rtype: axis.plot return
    """
    # Values
    y = values
    length = len(values)
    time = length * dt
    x = np.arange(t0, t0 + time, dt)
    x = x[: len(y)]
    # Axes
    add_axes(x, y, ax)
    # Plot
    obj = ax.plot(x, y, **kwargs)
    # Limits
    ax.set_xlim(t0, t0 + time)
    if (np.min(y) == 0) or (
        (np.min(y) > 0) and (np.min(y) - 0.01 * (np.max(y) - np.min(y)) < 0)
    ):
        ylims = ax.get_ylim()
        ylims = (0, ylims[1])
        ax.set_ylim(ylims)
    return obj


def plot_timeline_with_stds(
    y: np.ndarray,
    ax: plt.Axes,
    t0: float,
    dt: float,
    y_std: np.ndarray or None = None,
    name: str = "",
    add_label: bool = False,
    linewidth: float = 1,
    alpha: float = 0.2,
) -> Tuple[Tuple, str]:
    # Values
    length = len(y)
    time = length * dt
    x = np.arange(t0, t0 + time, dt)
    x = x[: len(y)]
    # Setup Plotting
    elements = []
    label = f"{latex_format(name)} "
    # Shadow
    if y_std is not None:
        e = ax.fill_between(x, y - y_std, y + y_std, alpha=alpha)
        elements.append(e)
        if add_label:
            label += " $\\mu \\pm \\sigma_\\mu$"
    # Timeline
    e = ax.plot(x, y, linestyle="-", linewidth=linewidth)
    elements.append(e[0])
    # Axes
    add_axes(x, y, ax)
    # Limits
    ax.set_xlim(t0, t0 + time)
    return tuple(elements), label


def plot_timetraces(
    file_path: str,
    out_path: str or None = None,
    properties: List = ("gamma_c", "gamma_n", "gamma_n_spectral"),
    # properties: List = ("energy", "kinetic_energy", "thermal_energy", "enstrophy", "enstrophy_phi"),
    t0: int = 0,
    t0_std: float = 300,
):
    with h5py.File(file_path, "r") as hf:
        parameters = dict(hf.attrs)
        fig, ax = plt.subplots(1, 1, figsize=(7, 3))
        t0_idx = int(t0 // parameters["dt"])
        max_len = 0
        for prop in properties:
            max_len = max(max_len, len(hf[prop]))
            if t0_idx >= len(hf[prop]):
                print(
                    f"{prop}: Not sufficient data. Range selected starts at t0={t0} to plot, data ends at t={len(hf[prop])*parameters['dt']:.2f}"
                )
                return
        t0_std_idx = int(t0_std // parameters["dt"])
        if t0_std_idx > max_len:
            print("WARNING start of statistics index is bigger than file length!")
            print(f"Calculating stats now from t0: {t0_std} -> {t0}")
            t0_std = t0
            t0_std_idx = t0_idx
        age = hf[list(hf.keys())[0]].shape[0] * parameters["dt"]
        elements = []
        labels = []
        min_yval = 0
        max_yval = 0
        for property in properties:
            prop_std = np.std(hf[property][t0_std_idx:])
            prop_mean = np.mean(hf[property][t0_std_idx:])
            prop_data = hf[property][t0_idx:]
            min_yval = min(min_yval, np.min(hf[property]))
            max_yval = max(max_yval, np.min(hf[property]))
            if not len(prop_data):
                print(f"WARNING no data in: {property}")
                continue
            element, label = plot_timeline_with_stds(
                prop_data,
                t0=t0,
                dt=parameters["dt"],
                ax=ax,
                name=property,
                add_label=False,
            )
            label += f" = {prop_mean:.2f}$\\pm${prop_std:.2f}"
            labels.append(label)
            elements.append(element[0])
        # Adjust y-axis
        if np.isclose(min_yval, 0, rtol=1e-05, atol=1e-08, equal_nan=False) or (
            (min_yval > 0) and (min_yval - 0.01 * (max_yval - min_yval) < 0)
        ):
            ylims = ax.get_ylim()
            ylims = (0, ylims[1])
            ax.set_ylim(ylims)
        # Adjust Labels
        ax.legend(elements, labels)
        ax.set_ylabel("value")
        ax.set_xlabel("time (t)")
        ax.xaxis.set_ticks(range(0, int(age) + 1, 100))
        fig.tight_layout()
        if out_path is None:
            out_path = (
                str(file_path).split(".h5")[0] + f"_{'-'.join(properties)}" + ".jpg"
            )
        fig.savefig(out_path)
        print(out_path)


def main(
    file_path: str,
    out_path: str or None = None,
    t0: int = 0,
    t0_std: float = 300,
):
    plot_timetraces(
        file_path=file_path,
        out_path=out_path,
        properties=("gamma_c", "gamma_n", "gamma_n_spectral"),
        t0=t0,
        t0_std=t0_std,
    )
    plot_timetraces(
        file_path=file_path,
        out_path=out_path,
        properties=("energy", "kinetic_energy", "thermal_energy", "enstrophy", "enstrophy_phi"),
        t0=t0,
        t0_std=t0_std,
    )

if __name__ == "__main__":
    fire.Fire(main)



from typing import Tuple
import numpy as np
import scipy.signal


def mean_downsample2x(x: np.ndarray):
    """mean of every 2 cells, repeatedly apply to reduce higher orders

    :param x: array, first axis ignored
    :type x: array
    :return: array of reduced dimensions by (None, 2, 2, ...)
    :rtype: array
    """
    rank = len(x.shape)
    for i in range(1, rank):  # Ignore first
        if x.shape[i] == 1:  # Until first 1 dimension
            break
        a = (slice(None),) * rank  # Pass everythin
        # Now replace the "appropriate"
        x = (
            x[tuple(a[:i] + (slice(0, None, 2),) + a[i + 1 :])]
            + x[tuple(a[:i] + (slice(1, None, 2),) + a[i + 1 :])]
        ) / 2
    return x


def skip_downsampleNx(x: np.ndarray, lowerscale: float) -> np.ndarray:
    """Take every Nth value, no interpolation performed!

    :param x: input array
    :type x: array
    :param lowerscale: factor by which it is downsampled along each axis
    :type lowerscale: integer
    :return: downsampled array
    :rtype: array
    """
    # Just step through by lowerscale
    slices = [slice(None)] + [
        slice(None, None, lowerscale) for _ in range(len(x.shape) - 1)
    ]
    x = x[tuple(slices)]
    return x


def median_downsampleNx(x: np.ndarray, lowerscale: float) -> np.ndarray:
    new_shape = [x.shape[0]] + list(np.array(x.shape[1:]) // lowerscale)
    # Along axis where the shapes are not equal
    along_axis = np.arange(len(x.shape))[np.not_equal(x.shape, x.shape)]
    # Insert axis to be reduced on alter on
    sh = x.shape
    for j, i in enumerate(along_axis):
        sh = np.insert(sh, i + 1 + j, x.shape[i] // new_shape[i])
    # Group excess into new (temporary) dimensions
    x = x.reshape(sh)
    # Apply gather function along temporary dimension in reshaped array
    for ax in along_axis:
        x = np.median(x, ax + 1)
    return x


def fourier_downsample(x: np.ndarray, lowerscale: float, axes: Tuple[int] = (-1, -2)) -> np.ndarray:
    """resample in fourier space and will round output shape"""
    rounded_shape = list(map(round, np.array(x.shape) / lowerscale))
    new_lengths = np.array(rounded_shape).astype(int)
    for axis_idx in axes:
        x = scipy.signal.resample(x, num=new_lengths[axis_idx], axis=axis_idx)
    return x
5


"""
hw2d.model: Hasegawa-Wakatani 2D Simulation Module

This module provides the core functionality to simulate the Hasegawa-Wakatani (HW) model in two dimensions. 
It offers flexibility in terms of numerical schemes and allows for comprehensive profiling and debugging 
of the simulation process.

Functions:
    - euler_step: Implements the Euler time-stepping method.
    - rk4_step: Implements the Runge-Kutta 4th order time-stepping method.
    - get_phi: Computes the electrostatic potential from vorticity.
    - diffuse: Applies diffusion to an array.
    - gradient_2d: Computes the 2D gradient of the system.
    - get_gammas: Computes energy gradients.

Classes:
    - HW: Represents the primary simulation entity for the Hasegawa-Wakatani model.

Notes:
    The module supports both NumPy and Numba for computational operations and provides detailed logging and 
    profiling capabilities.
"""
from typing import Tuple, Callable, Dict
import numpy as np
import matplotlib.pyplot as plt
import time
import pandas as pd

# Local Imports
from hw2d.initializations.fourier_noise import get_fft_noise
from hw2d.utils.namespaces import Namespace

# NumPy Version
from hw2d.gradients.numpy_gradients import periodic_laplace_N, periodic_gradient
from hw2d.poisson_bracket.numpy_arakawa import periodic_arakawa_vec
from hw2d.poisson_solvers.numpy_fourier_poisson import fourier_poisson_double
from hw2d.physical_properties.numpy_properties import *
periodic_arakawa = periodic_arakawa_vec

try:
    # Numba
    from hw2d.gradients.numba_gradients import periodic_laplace_N, periodic_gradient
    from hw2d.poisson_bracket.numba_arakawa import periodic_arakawa_stencil
    from hw2d.poisson_solvers.numba_fourier_poisson import fourier_poisson_double

    periodic_arakawa = periodic_arakawa_stencil
except:
    pass

# Other
# from hw2d.poisson_solvers.pyfftw_fourier_poisson import fourier_poisson_pyfftw


class HW:
    def __init__(
        self,
        dx: float,
        N: int,
        c1: float,
        nu: float,
        k0: float,
        poisson_bracket_coeff: float = 1,
        kappa_coeff: float = 1,
        debug: bool = False,
        TEST_CONSERVATION: bool = True,
    ):
        """
        Initialize the Hasegawa-Wakatani (HW) simulation.

        Parameters:
            dx (float): Grid spacing.
            N (int): System size.
            c1 (float): Model-specific parameter.
            nu (float): Diffusion coefficient.
            k0 (float): Fundamental wavenumber.
            poisson_bracket_coeff (float, optional): Coefficient for the Arakawa scheme. Default is 1.
            kappa_coeff (float, optional): Coefficient for d/dy phi. Default is 1.
            debug (bool, optional): Flag to enable debugging mode. Default is False.
            TEST_CONSERVATION (bool, optional): Flag to test conservation properties. Default is True.
        """
        # Numerical Schemes
        self.poisson_solver: Callable = fourier_poisson_double
        self.diffuse_N: Callable = periodic_laplace_N
        self.poisson_bracket: Callable = periodic_arakawa
        self.gradient_func: Callable = periodic_gradient
        # Physical Values
        self.N: int = int(N)
        self.c1: float = c1
        self.nu: float = (-1) ** (self.N + 1) * nu
        self.k0: float = k0
        self.poisson_bracket_coeff: float = poisson_bracket_coeff
        self.kappa_coeff: float = kappa_coeff
        self.dx: float = dx
        self.L: float = 2 * np.pi / k0
        # Physical Properties
        self.TEST_CONSERVATION: bool = TEST_CONSERVATION
        # Debug Values
        self.debug: bool = debug
        self.counter: int = 0
        self.watch_fncs: Tuple[str] = (
            "full_step",
            "get_phi",
            "diffuse",
            "gradient_2d",
            "poisson_bracket",
        )
        self.timings: Dict[str, float] = {k: 0 for k in self.watch_fncs}
        self.calls: Dict[str, int] = {k: 0 for k in self.watch_fncs}

    def log(self, name: str, time: float):
        """
        Log the time taken by a specific function.

        Parameters:
            name (str): Name of the function.
            time (float): Time taken by the function.
        """
        self.timings[name] += time
        self.calls[name] += 1

    def print_log(self):
        """Display the timing information for the functions profiled."""
        df = pd.DataFrame({"time": self.timings, "calls": self.calls})
        df["time/call"] = df["time"] / df["calls"]
        df["%time"] = df["time"] / df["time"]["full_step"] * 100
        df.sort_values("time/call", inplace=True)
        print(df)

    def euler_step(
        self, plasma: Namespace, dt: float, dx: float, **kwargs
    ) -> Namespace:
        t0 = time.time()
        y = Namespace(**{k:v for k,v in plasma.items() if k != "phi"})
        if "phi" in plasma:
            d = dt * self.gradient_2d(**y, phi=plasma["phi"], dt=0, dx=dx)
        else:
            d = dt * self.gradient_2d(**y, dt=0, dx=dx)
        y += d
        y["phi"] = self.get_phi(omega=y.omega, dx=dx)
        y["age"] = plasma.age + dt
        # Wrap-up
        self.log("full_step", time.time() - t0)
        return y

    def rk4_step(self, plasma: Namespace, dt: float, dx: float, **kwargs) -> Namespace:
        # RK4
        t0 = time.time()
        yn = Namespace(**{k:v for k,v in plasma.items() if k != "phi"})
        # pn = self.get_phi(omega=yn.omega, dx=dx)  # TODO: only execute for t=0
        pn = plasma.phi
        k1 = dt * self.gradient_2d(**yn, phi=pn, dt=0, dx=dx)
        y1 = yn + k1 * 0.5
        p1 = self.get_phi(omega=y1.omega, dx=dx)
        k2 = dt * self.gradient_2d(**y1, phi=p1, dt=dt / 2, dx=dx)
        y2 = yn + k2 * 0.5
        p2 = self.get_phi(omega=y2.omega, dx=dx)
        k3 = dt * self.gradient_2d(**y2, phi=p2, dt=dt / 2, dx=dx)
        y3 = yn + k3
        p3 = self.get_phi(omega=y3.omega, dx=dx)
        k4 = dt * self.gradient_2d(**y3, phi=p3, dt=dt, dx=dx)
        # p4 = self.get_phi(k4.omega)
        # TODO: currently adds two timesteps
        yk1 = yn + (k1 + 2 * k2 + 2 * k3 + k4) * (1 / 6)
        phi = self.get_phi(omega=y1.omega, dx=dx)
        # Set properties not valid through y1
        yk1["phi"] = phi
        yk1["age"] = plasma.age + dt
        # Wrap-up
        self.log("full_step", time.time() - t0)
        if self.debug:
            print(
                " | ".join(
                    [
                        f"{plasma.age + dt:<7.04g}",
                        f"{np.max(np.abs(yn.density.data)):>7.02g}",
                        f"{np.max(np.abs(k1.density.data)):>7.02g}",
                        f"{np.max(np.abs(k2.density.data)):>7.02g}",
                        f"{np.max(np.abs(k3.density.data)):>7.02g}",
                        f"{np.max(np.abs(k4.density.data)):>7.02g}",
                        f"{time.time()-t0:>6.02f}s",
                    ]
                )
            )
        return yk1

    def get_phi(self, omega: np.ndarray, dx: float) -> np.ndarray:
        t0 = time.time()
        o_mean = np.mean(omega)
        centered_omega = omega - o_mean
        phi = self.poisson_solver(tensor=centered_omega, dx=dx)
        self.log("get_phi", time.time() - t0)
        return phi

    def diffuse(self, arr: np.ndarray, dx: float, N: int) -> np.ndarray:
        t0 = time.time()
        arr = self.diffuse_N(arr=arr, dx=dx, N=N)
        self.log("diffuse", time.time() - t0)
        return arr

    def gradient_2d(
        self,
        density: np.ndarray,
        omega: np.ndarray,
        dt: float,
        dx: float,
        phi: np.ndarray or None = None,
        debug: bool = False,
        **kwargs
    ) -> Namespace:
        if phi is None:
            phi = self.get_phi(omega, dx=dx)
        # Setup
        arak_comp_o = 0
        arak_comp_n = 0
        kap = 0
        Do = 0
        Dn = 0
        t0 = time.time()
        dy_p = self.gradient_func(phi, dx=dx, axis=0)
        self.log("gradient_2d", time.time() - t0)

        # Calculate Gradients
        diff = phi - density

        # Step 2.1: New Omega.
        o = self.c1 * diff
        if self.poisson_bracket_coeff:
            t0 = time.time()
            arak_comp_o = -self.poisson_bracket_coeff * self.poisson_bracket(
                zeta=phi, psi=omega, dx=dx
            )
            self.log("poisson_bracket", time.time() - t0)
            o += arak_comp_o
        if self.nu:
            Do = self.nu * self.diffuse(arr=omega, dx=dx, N=self.N)
            o += Do

        # Step 2.2: New Density.
        n = self.c1 * diff
        if self.poisson_bracket_coeff:
            t0 = time.time()
            arak_comp_n = -self.poisson_bracket_coeff * self.poisson_bracket(
                zeta=phi, psi=density, dx=dx
            )
            self.log("poisson_bracket", time.time() - t0)
            n += arak_comp_n
        if self.kappa_coeff:
            kap = -self.kappa_coeff * dy_p
            n += kap
        if self.nu:
            Dn = self.nu * self.diffuse(arr=density, dx=dx, N=self.N)
            n += Dn

        # Print gradients to see which one explodes
        if debug:
            print(
                "  |  ".join(
                    [
                        f"  dO/dt = {np.max(np.abs(self.c1 * diff)):>8.2g} + {np.max(np.abs(arak_comp_o)):>8.2g} + {np.max(np.abs(Do)):>8.2g}"
                        f"  dn/dt = {np.max(np.abs(self.c1 * diff)):>8.2g} + {np.max(np.abs(arak_comp_n)):>8.2g} + {np.max(np.abs(kap)):>8.2g} + {np.max(np.abs(Dn)):>8.2g}",
                        f"  dO/dt = {np.mean(self.c1 * diff):>8.2g} + {np.mean(arak_comp_o):>8.2g} + {np.mean(Do):>8.2g}",
                        f"  dn/dt = {np.mean(self.c1 * diff):>8.2g} + {np.mean(arak_comp_n):>8.2g} + {np.mean(kap):>8.2g} + {np.mean(Dn):>8.2g}",
                    ]
                )
            )

        # Structure return
        return_dict = Namespace(
            density=n,
            omega=o
        )

        if "age" in kwargs:
            return_dict["age"] = kwargs["age"] + dt

        # Add energy gradients for testing conservation
        if self.TEST_CONSERVATION:
            gamma_n, gamma_c = self.get_gammas(density, phi)
            Dp = self.nu * self.diffuse(arr=phi, dx=dx, N=self.N)
            DE = get_DE(n=density, p=phi, Dn=Dn, Dp=Dp)
            DU = get_DU(n=density, o=omega, Dn=Dn, Dp=Dp)
            dE_dt = get_dE_dt(gamma_n=gamma_n, gamma_c=gamma_c, DE=DE)
            dU_dt = get_dU_dt(gamma_n=gamma_n, DU=DU)
            return_dict["dE"] = dE_dt
            return_dict["dU"] = dU_dt

        return return_dict

    def get_gammas(self, n: np.ndarray, p: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        # Generate Energy Gradients
        gamma_n = get_gamma_n(n=n, p=p, dx=self.dx)
        gamma_c = get_gamma_c(n=n, p=p, c1=self.c1, dx=self.dx)
        return gamma_n, gamma_c



"""
NumPy-based Gradient Computation
================================

This module offers a collection of functions for computing gradients on 2D grids using the NumPy library. 
It provides a standard implementation based on array computation suitable for solving the HW2D model, including:

- Basic Gradient Computation (`gradient`): Computes the gradient of a 2D array using central finite differences.
- Periodic Gradient (`periodic_gradient`): Computes the gradient with periodic boundary conditions.
- Laplace Operations:
    - Basic Laplace Computation (`laplace`): Computes the Laplace using finite differences.
    - Periodic Laplace (`periodic_laplace`): Laplace operation with periodic boundary conditions.
    - Iterative Laplace (`periodic_laplace_N`): Computes the Laplace N times successively.
    - Fourier-based Laplace (`fourier_laplace`): Computes the Laplace using Fourier transforms for enhanced accuracy.

All functions in this module are optimized for performance while ensuring accuracy, making them suitable for both prototyping and production-level simulations.
"""

import numpy as np


def laplace(padded: np.ndarray, dx: float):
    """
    Compute the Laplace of a 2D array using finite differences.

    Args:
        padded (np.ndarray): 2D array with padding of size 1.
        dx (float): The spacing between grid points.

    Returns:
        np.ndarray: The Laplace of the input array.
    """
    return (
        padded[..., 0:-2, 1:-1]  # above
        + padded[..., 1:-1, 0:-2]  # left
        - 4 * padded[..., 1:-1, 1:-1]  # center
        + padded[..., 1:-1, 2:]  # right
        + padded[..., 2:, 1:-1]  # below
    ) / dx**2


def periodic_laplace(arr: np.ndarray, dx: float):
    """
    Compute the Laplace of a 2D array using finite differences with periodic boundary conditions.

    Args:
        a (np.ndarray): Input 2D array.
        dx (float): The spacing between grid points.

    Returns:
        np.ndarray: The Laplace of the input array with periodic boundary conditions.
    """
    pad_size = 1
    if len(arr.shape) > 2:
        pad_size = [(0, 0) for _ in range(len(arr.shape))]
        pad_size[-1] = (1, 1)
        pad_size[-2] = (1, 1)
    return laplace(np.pad(arr, pad_size, "wrap"), dx)


def periodic_laplace_N(arr: np.ndarray, dx: float, N: int) -> np.ndarray:
    """
    Compute the Laplace of a 2D array using finite differences N times successively with periodic boundary conditions.

    Args:
        a (np.ndarray): Input 2D array.
        dx (float): The spacing between grid points.
        N (int): Number of iterations.

    Returns:
        np.ndarray: The Laplace of the input array with periodic boundary conditions.
    """
    for _ in range(N):
        arr = periodic_laplace(arr, dx)
    return arr


def fourier_laplace(grid: np.ndarray, dx: float, times: int = 1) -> np.ndarray:
    """
    Compute the Laplace of a 2D array using Fourier transform.

    Args:
        grid (np.ndarray): Input 2D array.
        dx (float): The spacing between grid points.
        times (int, optional): Number of times to apply the Laplace operator. Default is 1.

    Returns:
        np.ndarray: The Laplace of the input array using Fourier transform.
    """
    frequencies = np.fft.fft2(grid.astype(np.complex128))
    k = np.meshgrid(*[np.fft.fftfreq(int(n)) for n in grid.shape], indexing="ij")
    k = np.expand_dims(np.stack(k, -1), 0)
    k = k.astype(np.float64)
    k_squared = np.sum(k**2, axis=-1)
    fft_laplace = -((2 * np.pi) ** 2) * k_squared
    result = np.real(np.fft.ifft2(frequencies * fft_laplace**times))
    return (result / dx**2).astype(grid.dtype)[0]


def gradient(padded: np.ndarray, dx: float, axis: int = 0) -> np.ndarray:
    """
    Compute the gradient of a 2D array using finite differences.

    Args:
        padded (np.ndarray): 2D array with padding of size 1.
        dx (float): The spacing between grid points.
        axis (int): Axis along which the gradient is tkaen

    Returns:
        np.ndarray: Gradient in axis-direction.
    """
    if axis == 0:
        return (padded[2:, 1:-1] - padded[0:-2, 1:-1]) / (2 * dx)
    elif axis == 1:
        return (padded[1:-1, 2:] - padded[1:-1, 0:-2]) / (2 * dx)
    elif axis == -2:
        return (padded[..., 2:, 1:-1] - padded[..., 0:-2, 1:-1]) / (2 * dx)
    elif axis == -1:
        return (padded[..., 1:-1, 2:] - padded[..., 1:-1, 0:-2]) / (2 * dx)


def periodic_gradient(input_field: np.ndarray, dx: float, axis: int = 0) -> np.ndarray:
    """
    Compute the gradient of a 2D array using finite differences with periodic boundary conditions.

    Args:
        input_field (np.ndarray): Input 2D array.
        dx (float): The spacing between grid points.
        axis (int): Axis along which the gradient is tkaen

    Returns:
        tuple: Gradient in y-direction, gradient in x-direction with periodic boundary conditions.
    """
    if axis < 0:
        pad_size = [(0, 0) for _ in range(len(input_field.shape))]
        pad_size[-1] = (1, 1)
        pad_size[-2] = (1, 1)
    else:
        pad_size = 1
    padded = np.pad(input_field, pad_width=pad_size, mode="wrap")
    return gradient(padded, dx, axis=axis)



import pytest
import h5py
import numpy as np
import os
from hw2d.physical_properties.numpy_properties import *


reference = h5py.File(
    f"{os.path.dirname(os.path.realpath(__file__))}/reference.h5", "r"
)
n = reference["density"][0]
p = reference["phi"][0]
o = reference["omega"][0]
dx = reference.attrs["dx"]
c1 = reference.attrs["c1"]


def test_gamma_n():
    gamma_n = get_gamma_n(n=n, p=p, dx=dx)
    assert np.isclose(gamma_n, 0.57292056), gamma_n


def test_gamma_n_spectrally():
    gamma_n_spectrally = get_gamma_n_spectrally(n=n, p=p, dx=dx)
    assert np.isclose(gamma_n_spectrally, 0.5745075066507911), gamma_n_spectrally


def test_gamma_c():
    gamma_c = get_gamma_c(n=n, p=p, c1=c1, dx=dx)
    assert np.isclose(gamma_c, 0.55321026), gamma_c


def test_energy():
    energy = get_energy(n=n, phi=p, dx=dx)
    assert np.isclose(energy, 3.7285664081573486), energy


def test_energy_N():
    energy_N = get_energy_N_spectrally(n=n)
    assert np.isclose(energy_N, 2.096125), energy_N


def test_energy_V():
    energy_V = get_energy_V_spectrally(p=p, dx=dx)
    assert np.isclose(energy_V, 1.6418878080857098), energy_V


def test_enstrophy():
    enstrophy = get_enstrophy(n=n, omega=o, dx=dx)
    assert np.isclose(enstrophy, 11.743794), enstrophy


def test_enstrophy_phi():
    enstrophy_phi = get_enstrophy_phi(n=n, phi=p, dx=dx)
    assert np.isclose(enstrophy_phi, 11.453048), enstrophy_phi


if __name__ == "__main__":
    pytest.main()



"""
The `hw2d.poisson_solvers` Module
=================================

This module provides a suite of solvers for the Poisson equation, which is a fundamental equation in fluid dynamics and numerical simulations. Given the gradient (or Laplace) of a field, these solvers can retrieve the original field. They are crucial for inverse operations in spectral methods and other numerical techniques.

The solvers are organized based on their computational framework and method:

- **NumPy**: Solvers that leverage the NumPy library for efficient array operations.
- **Numba**: Solvers that utilize the Numba JIT compiler for enhanced performance.
- **Fourier-based**: Solvers that leverage the Fourier transform for solving the Poisson equation in spectral space.

To explore the specific functions and their documentation, navigate to the respective sub-modules:

- `hw2d.poisson_solvers.numpy_fourier_poisson`: Fourier-based Poisson solvers using NumPy.
- `hw2d.poisson_solvers.numba_fourier_poisson`: Fourier-based Poisson solvers optimized with Numba.

Select the appropriate solver based on your application's requirements and available computational resources.
"""



names = {
    "int_gamma_n": "$\\Gamma_n(t) = \\int{dk_y \\; \\Gamma_n \\left( k_y \\right) }$",
    "integrated gamma_n": "$\\Gamma_n(t) = \\int{dk_y \\; \\Gamma_n \\left( k_y \\right) }$",
    "gamma_n": "$\\Gamma_n$",
    "gamma_n_k": "$\\Gamma_n \\left(k_y\\right)$",
    "gamma_n_spectral": "$\\int{\\Gamma_n \\left( k_y \\right) }$",
    "gamma_c": "$\\Gamma_c$",
    "ky_phi": "$\\phi \\left( k_y \\right)$",
    "ky_omega": "$\\Omega \\left( k_y \\right)$",
    "ky_density": "$n \\left( k_y \\right)$",
    "ky": "$k_y$",
    "mean": "$\\mu$",
    "std": "$\\sigma$",
    "delta_k": "$\\delta_k$",
    "U_phi": "$U_\\phi$",
    "U_centered": "$U \\left( \\Omega-\\mu \\left( \\Omega \\right), n \\right)$",
    "nu": "$\\nu$",
    "density": "density ($n$)",
    "phi": "potential ($\\phi$)",
    "omega": "vorticity ($\\Omega$)",
    "dE_dt": "$\\partial_t E$",
    "dU_dt": "$\\partial_t U$",
    "DE_v": "$DE_v$",
    "DE_n": "$DE_n$",
    "DEv": "$DE_v$",
    "DEn": "$DE_n$",
    "DE": "$DE$",
    "DU": "$DU$",
    "E": "E",
    "U": "U",
    "cfl": "cfl",
    "mean_gamma_n": "$\\mu \\left( \\Gamma_n \\right)$",
    "std_gamma_n": "$\\sigma \\left( \\Gamma_n \\right)$",
    "mean_gamma_c": "$\\mu \\left( \\Gamma_c \\right)$",
    "std_gamma_c": "$\\sigma \\left( \\Gamma_c \\right)$",
    "mean_int_gamma_n": "$\\mu_t \\left( \\int \\Gamma_n \\left(k_y\\right) \\right)$",
    "std_gamma_n": "$\\sigma \\left( \\int \\Gamma_n \\left(k_y\\right) \\right)$",
    "c1": "$c_1$",
    "kinetic_energy": "$E_K$",
    "thermal_energy": "$E_T$",
    "energy": "Energy $(E)$",
    "enstrophy": "Enstrophy $(U)$",
    "enstrophy_phi": "Enstrophy $(U(n,\\phi))$",
}


def metric_format(string: str) -> str:
    if any([s in string for s in ["mean", "median", "std"]]):
        return string[: string.find("_")] + " " + names[string[string.find("_") + 1 :]]
    return ""


def latex_format(string: str) -> str:
    try:
        return names[string]
    except Exception as e:
        print(string, e)
        return string



"""
`fourier_noise.py`: A module to generate 2D noise patterns using the Fast Fourier Transform (FFT).

This module provides functionality to generate 2D noise patterns based on Fourier space manipulation.
The generated noise can be controlled by specifying the desired resolution, size, scale, frequency range, 
and wavelength range. The resulting noise can be useful for various applications in simulations, testing,
and visualization.
"""
import numpy as np
from typing import Tuple


def get_fft_noise(
    resolution: Tuple[int, int],
    size: float,
    scale: float,
    min_frequency: float = 0,
    max_frequency: float = 0,
    min_wavelength: float = 0,
    max_wavelength: float = 0,
    factor: int = 2,
) -> np.ndarray:
    """
    Generate a 2D noise pattern using the FFT.

    This function creates a 2D noise pattern based on standard normal distributions in Fourier space.
    It allows for control over the frequency and wavelength components of the noise through the provided parameters.

    Args:
        resolution (Tuple[int, int]): The dimensions of the generated noise.
        size (float): Physical size.
        scale (float): Scaling factor for the frequencies.
        min_frequency (float, optional): Minimum frequency for filtering. Default is 0.
        max_frequency (float, optional): Maximum frequency for filtering. Default is 0.
        min_wavelength (float, optional): Minimum wavelength for filtering. Default is 0.
        max_wavelength (float, optional): Maximum wavelength for filtering. Default is 0.
        factor (int, optional): Factor used in calculating the frequency components. Default is 2.

    Returns:
        np.ndarray: The generated 2D noise pattern.
    """
    # Calculate random complex values
    shape = (1, *resolution, 1)
    rnd_real = np.random.standard_normal(shape).astype(np.complex128)
    rnd_imag = 1j * np.random.standard_normal(shape).astype(np.complex128)
    rndj = rnd_real + rnd_imag

    # Calculate frequency components
    k = np.meshgrid(*[np.fft.fftfreq(int(n)) for n in resolution], indexing="ij")
    k = np.expand_dims(np.stack(k, -1), 0)
    k = k * resolution / size * scale  # in physical units
    k = np.sum(np.abs(k) ** factor, axis=-1, keepdims=True)

    # Convert wavelengths to frequencies if provided
    if max_wavelength:
        min_frequency = 1 / max_wavelength
    if min_wavelength:
        max_frequency = 1 / min_wavelength

    # Create frequency mask
    weight_mask = np.ones(shape)
    if min_frequency:
        weight_mask += 1 / (1 + np.exp((min_frequency - k) * 1e3)) - 1
    if max_frequency:
        with np.errstate(divide="ignore", invalid="ignore"):
            weight_mask -= 1 / (1 + np.exp((max_frequency - k) * 1e3))
    # Check weight mask
    assert np.all(weight_mask <= 1) and np.all(
        weight_mask >= 0
    ), "Weight mask values out of bounds."

    # Handle division by zero for k
    k[(0,) * len(k.shape)] = np.inf
    inv_k = 1 / k
    inv_k[(0,) * len(k.shape)] = 0

    # Compute result
    smoothness = 1
    fft = rndj * inv_k**smoothness * weight_mask
    array = np.real(np.fft.ifft2(fft, axes=[1, 2]))
    array /= np.std(array, axis=tuple(range(1, len(array.shape))), keepdims=True)
    array -= np.mean(array, axis=tuple(range(1, len(array.shape))), keepdims=True)
    array = array.astype(np.float64)
    return array[0, ..., 0]



from unittest import TestCase

import numpy as np
from hw2d.poisson_bracket.numba_arakawa import periodic_arakawa_stencil as periodic_arakawa_nb
from hw2d.poisson_bracket.numba_arakawa import periodic_arakawa_vec as periodic_arakawa_vec_nb
from hw2d.poisson_bracket.numpy_arakawa import periodic_arakawa as periodic_arakawa
from hw2d.poisson_bracket.numpy_arakawa import periodic_arakawa_vec as periodic_arakawa_np
from hw2d.gradients.numpy_gradients import periodic_gradient
from hw2d.initializations.sine import get_2d_sine


class test_arakawa(TestCase):
    def test_periodic_arakawa_nb(self):
        N = 128
        grid_size = (N, N)
        L = 1
        dx = 0.2
        input_field = get_2d_sine(grid_size, L)
        zeta = input_field.copy()
        psi = input_field.copy() / -2.0

        zeta = np.random.rand(N, N)
        psi = zeta.T #np.random.rand(N, N) / -2.0

        arr1 = periodic_arakawa(zeta, psi, dx)
        arr2 = periodic_arakawa_np(zeta, psi, dx)
        arr3 = periodic_arakawa_nb(zeta, psi, dx)
        arr4 = periodic_arakawa_vec_nb(zeta, psi, dx)
        assert np.allclose(arr1, arr2)
        assert np.allclose(arr1, arr3)
        assert np.allclose(arr1, arr4)


if __name__ == "__main__":
    test_arakawa().test_periodic_arakawa_nb()



"""
The `hw2d.physical_properties` Module
=====================================

This module offers the computation of standard values for physical properties of the HW2D system.
These properties are essential for diagnostics, understanding system dynamics, and providing insights into the behavior of fluid simulations.
They include scalar values, spectral properties, and "conservation" laws.

The properties are computed using various methods:

- **NumPy**: Efficient array operations leveraging the NumPy library.
- **Numba**: JIT-compiled functions for performance-critical operations.

To explore specific functionalities:

- `hw2d.physical_properties.numpy_properties`: Properties computed using NumPy.
- `hw2d.physical_properties.numba_properties`: Properties optimized with Numba.

For detailed usage, navigate to the respective sub-modules and refer to their documentation.
"""



"""
The `hw2d.gradients` Module
===========================

This module contains a comprehensive set of functions for computing gradients, an essential operation in fluid dynamics and numerical simulations. 
These gradient functions facilitate various operations including:

- Laplace operations: Compute the second derivative of a field.
- Finite difference schemes: Approximate derivatives of functions using discrete points.
- Fourier-based gradient calculations: Use Fourier transforms for efficient and accurate gradient computations.

Implementations are organized based on the computational framework used:

- **NumPy**: Functions that leverage the NumPy library for efficient array operations.
- **Numba**: Functions that utilize the Numba JIT compiler for enhanced performance, especially suitable for large-scale simulations.

To see specific functions and their documentation, navigate to the respective sub-modules:

- `hw2d.gradients.numpy_gradients`: Gradient functions implemented using NumPy.
- `hw2d.gradients.numba_gradients`: Gradient functions accelerated with Numba.

Remember to choose the appropriate implementation based on your application's needs and the computational resources available.
"""



"""
The `hw2d.poisson_bracket` Module
=========================

This module provides implementations of the Arakawa scheme to compute the Poisson bracket. 
The Arakawa scheme ensures accurate computation of the Poisson bracket, avoiding spurious oscillations which can arise from naive discretization.

The module offers implementations based on different computational frameworks:

- **NumPy**: Efficient array operations and integration with the scientific Python ecosystem.
- **Numba**: Just-in-time compiled functions optimized for speed.

To see specific functions and their documentation, navigate to the respective sub-modules:

- `hw2d.poisson_bracket.numpy_arakawa`: Arakawa scheme functions implemented using NumPy.
- `hw2d.poisson_bracket.numba_arakawa`: Arakawa scheme functions accelerated with Numba.
"""

# # arakawa_scheme/__init__.py

# from . import numpy_arakawa

# try:
#     from . import numba_arakawa
#     _NUMBA_AVAILABLE = True
#     _USE_NUMBA = True  # Default to use Numba if available
# except ImportError:
#     _NUMBA_AVAILABLE = False
#     _USE_NUMBA = False  # Fallback to not using Numba

# def use_numba(enable=True):
#     """Enable or disable Numba-accelerated functions for Arakawa scheme."""
#     global _USE_NUMBA
#     if enable and not _NUMBA_AVAILABLE:
#         raise ImportError("Numba is not available on this system for Arakawa scheme.")
#     _USE_NUMBA = enable

# def periodic_arakawa(*args, **kwargs):
#     if _USE_NUMBA:
#         return numba_arakawa.periodic_arakawa_st2(*args, **kwargs)
#     else:
#         return numpy_arakawa.periodic_arakawa_vec(*args, **kwargs)



from typing import Dict, List, Tuple, Any
import h5py
import numpy as np
from hw2d.utils.namespaces import Namespace


def get_save_params(
    params: Dict[str, Any], dt: float, snaps: int, x: int, y: int, x_save: int or None = None, y_save: int or None = None
) -> Dict[str, Any]:
    params = params.copy()
    params["dt"] = dt
    params["frame_dt"] = dt * snaps
    params["x"] = x
    params["y"] = y
    params["x_save"] = x_save if x_save is not None else x
    params["y_save"] = y_save if y_save is not None else y
    params["grid_pts"] = x
    return params


def create_appendable_h5(
    filepath: str,
    params: Dict[str, Any],
    dtype: np.dtype = np.float32,
    chunk_size: int = 100,
    field_list: List[str] = ["density", "omega", "phi"],
) -> None:
    y = params["y_save"]
    x = params["x_save"]
    with h5py.File(f"{filepath}", "w") as hf:
        for field_name in field_list:
            hf.create_dataset(
                field_name,
                dtype=dtype,
                shape=(0, y, x),
                maxshape=(None, y, x),
                chunks=(chunk_size, y, x),
                compression="gzip",
            )
        for key, value in params.items():
            hf.attrs[key] = value


def append_h5(output_path: str, buffer: np.ndarray, buffer_index: int) -> None:
    """append a file, from buffer, with buffer_index size"""
    with h5py.File(output_path, "a") as hf:
        for field_name in buffer.keys():
            _ = hf[field_name].resize((hf[field_name].shape[0] + buffer_index), axis=0)
            hf[field_name][-buffer_index:] = buffer[field_name][:buffer_index]


def save_to_buffered_h5(
    buffer: Dict[str, Any],
    buffer_length: int,
    buffer_index: int,
    new_val: Dict[str, Any],
    output_path: str,
    field_list: List[str] = ["density", "omega", "phi"],
) -> int:
    """
    Save data to a buffer. If the buffer is full, flush the buffer to the HDF5 file.

    Args:
        buffer (Dict[str, Any]): Data buffer.
        buffer_length (int): Maximum size of the buffer.
        new_val (Dict[str, Any]): New values to be added to the buffer.
        field_list (List[str]): List of fields to be saved.
        buffer_index (int): Current index in the buffer.
        flush_index (int): Index to start flushing in the HDF5 file.
        output_path (str): Path of the output HDF5 file.

    Returns:
        Tuple[int, int]: Updated buffer index and flush index.
    """
    for idx, field in enumerate(field_list):
        buffer[field][buffer_index] = new_val[field]
    buffer_index += 1
    # If buffer is full, flush to HDF5 and reset buffer index
    if buffer_index == buffer_length:
        append_h5(output_path, buffer, buffer_index)
        buffer_index = 0
    return buffer_index


def create_fixed_h5(
    file_path: str,
    time: int,
    y: int,
    x: int,
    chunk_size: int = 100,
    compression: str = "gzip",
    field_list: List = ["density", "omega", "phi"],
    dtype=np.float32,
) -> None:
    with h5py.File(file_path, "w") as hf:
        for field_name in field_list:
            hf.create_dataset(
                field_name,
                dtype=dtype,
                shape=(time, y, x),
                maxshape=(None, y, x),
                chunks=(chunk_size, y, x),
                compression=compression,
            )


def load_h5_data(
    file_name: str, field_list: List[str]
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """
    Load data and attributes from an HDF5 file.

    Args:
        file_name (str): Name and path of the HDF5 file to load.
        field_list (List[str]): List of fields to be loaded.

    Returns:
        Tuple[Dict[str, Any], Dict[str, Any]]: Loaded data and associated parameters (attributes).
    """
    with h5py.File(file_name, "r") as h5_file:
        data = {}
        for field in field_list:
            data[field] = h5_file[field][:]
        params = dict(h5_file.attrs)
    return data, params


def continue_h5_file(
    file_name: str, field_list: List[str]
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """
    Load data and attributes from an HDF5 file.

    Args:
        file_name (str): Name and path of the HDF5 file to load.
        field_list (List[str]): List of fields to be loaded.

    Returns:
        Tuple[Dict[str, Any], Dict[str, Any]]: Loaded data and associated parameters (attributes).
    """
    lengths = []
    with h5py.File(file_name, "r") as h5_file:
        data = {}
        for field in field_list:
            data[field] = h5_file[field][-1].astype(np.float64)
            lengths.append(len(h5_file[field]))
        params = dict(h5_file.attrs)
    length = min(lengths)
    age = params["frame_dt"] * (length - 1)
    data = Namespace(**data, age=age, dx=params["dx"])
    params = {
        k: params[k]
        for k in ("dx", "N", "c1", "nu", "k0", "poisson_bracket_coeff", "kappa_coeff")
    }
    return data, params
